# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Plants
"""

from wuttafarm.db.model import EquipmentType, EquipmentAsset
from wuttafarm.web.views import TaxonomyMasterView
from wuttafarm.web.views.assets import AssetMasterView
from wuttafarm.web.forms.schema import EquipmentTypeRefs


class EquipmentTypeView(TaxonomyMasterView):
    """
    Master view for Equipment Types
    """

    model_class = EquipmentType
    route_prefix = "equipment_types"
    url_prefix = "/equipment-types"

    farmos_route_prefix = "farmos_equipment_types"
    farmos_entity_type = "taxonomy_term"
    farmos_bundle = "equipment_type"
    farmos_refurl_path = "/admin/structure/taxonomy/manage/equipment_type/overview"


class EquipmentAssetView(AssetMasterView):
    """
    Master view for Equipment Assets
    """

    model_class = EquipmentAsset
    route_prefix = "equipment_assets"
    url_prefix = "/assets/equipment"

    farmos_bundle = "equipment"
    farmos_refurl_path = "/assets/equipment"

    labels = {
        "equipment_types": "Equipment Type",
    }

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        equipment = f.model_instance

        # equipment_types
        f.fields.insert_after("asset_name", "equipment_types")
        f.set_node("equipment_types", EquipmentTypeRefs(self.request))
        if not self.creating:
            # nb. must explcitly declare value for non-standard field
            f.set_default("equipment_types", equipment.equipment_types)

        # manufacturer
        f.fields.insert_after("equipment_types", "manufacturer")

        # model
        f.fields.insert_after("manufacturer", "model")

        # serial_number
        f.fields.insert_after("model", "serial_number")

    def objectify(self, form):
        equipment = super().objectify(form)
        data = form.validated

        self.set_equipment_types(equipment, data["equipment_types"])

        return equipment

    def set_equipment_types(self, equipment, desired):
        model = self.app.model
        session = self.Session()
        current = [str(etype.uuid) for etype in equipment.equipment_types]

        for etype in desired:
            if etype["uuid"] not in current:
                equipment_type = session.get(model.EquipmentType, etype["uuid"])
                assert equipment_type
                equipment.equipment_types.append(equipment_type)

        desired = [etype["uuid"] for etype in desired]
        for uuid in current:
            if uuid not in desired:
                equipment_type = session.get(model.EquipmentType, uuid)
                assert equipment_type
                equipment.equipment_types.remove(equipment_type)


def defaults(config, **kwargs):
    base = globals()

    EquipmentTypeView = kwargs.get("EquipmentTypeView", base["EquipmentTypeView"])
    EquipmentTypeView.defaults(config)

    EquipmentAssetView = kwargs.get("EquipmentAssetView", base["EquipmentAssetView"])
    EquipmentAssetView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Animals
"""

from collections import OrderedDict

from webhelpers2.html import tags

from wuttaweb.forms.schema import WuttaDictEnum
from wuttaweb.forms.widgets import WuttaDateTimeWidget
from wuttaweb.util import get_form_data

from wuttafarm.db.model import AnimalType, AnimalAsset
from wuttafarm.web.views.assets import AssetTypeMasterView, AssetMasterView
from wuttafarm.web.forms.schema import AnimalTypeRef
from wuttafarm.web.forms.widgets import ImageWidget
from wuttafarm.web.util import get_farmos_client_for_user


class AnimalTypeView(AssetTypeMasterView):
    """
    Master view for Animal Types
    """

    model_class = AnimalType
    route_prefix = "animal_types"
    url_prefix = "/animal-types"

    farmos_entity_type = "taxonomy_term"
    farmos_bundle = "animal_type"
    farmos_refurl_path = "/admin/structure/taxonomy/manage/animal_type/overview"

    grid_columns = [
        "name",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "drupal_id",
        "farmos_uuid",
    ]

    has_rows = True
    row_model_class = AnimalAsset
    rows_viewable = True

    row_grid_columns = [
        "asset_name",
        "sex",
        "is_sterile",
        "birthdate",
        "archived",
    ]

    rows_sort_defaults = "asset_name"

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_farmos_url(self, animal_type):
        return self.app.get_farmos_url(f"/taxonomy/term/{animal_type.drupal_id}")

    def get_xref_buttons(self, animal_type):
        buttons = super().get_xref_buttons(animal_type)

        if animal_type.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_animal_types.view", uuid=animal_type.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons

    def delete(self):
        animal_type = self.get_instance()

        if animal_type.animal_assets:
            self.request.session.flash(
                "Cannot delete animal type which is still referenced by animal assets.",
                "warning",
            )
            url = self.get_action_url("view", animal_type)
            return self.redirect(self.request.get_referrer(default=url))

        return super().delete()

    def get_row_grid_data(self, animal_type):
        model = self.app.model
        session = self.Session()
        return (
            session.query(model.AnimalAsset)
            .join(model.Asset)
            .filter(model.AnimalAsset.animal_type == animal_type)
        )

    def configure_row_grid(self, grid):
        g = grid
        super().configure_row_grid(g)
        model = self.app.model
        enum = self.app.enum

        # asset_name
        g.set_link("asset_name")
        g.set_sorter("asset_name", model.Asset.asset_name)
        g.set_filter("asset_name", model.Asset.asset_name)

        # sex
        g.set_enum("sex", enum.ANIMAL_SEX)
        g.filters["sex"].verbs = ["equal", "not_equal"]

        # archived
        g.set_renderer("archived", "boolean")
        g.set_sorter("archived", model.Asset.archived)
        g.set_filter("archived", model.Asset.archived)

    def get_row_action_url_view(self, animal, i):
        return self.request.route_url("animal_assets.view", uuid=animal.uuid)

    def ajax_create(self):
        """
        AJAX view to create a new animal type.
        """
        model = self.app.model
        session = self.Session()
        data = get_form_data(self.request)

        name = data.get("name")
        if not name:
            return {"error": "Name is required"}

        animal_type = model.AnimalType(name=name)
        session.add(animal_type)
        session.flush()

        if self.app.is_farmos_mirror():
            client = get_farmos_client_for_user(self.request)
            self.app.auto_sync_to_farmos(animal_type, client=client)

        return {
            "uuid": animal_type.uuid.hex,
            "name": animal_type.name,
            "farmos_uuid": animal_type.farmos_uuid.hex,
            "drupal_id": animal_type.drupal_id,
        }

    @classmethod
    def defaults(cls, config):
        """ """
        cls._defaults(config)
        cls._animal_type_defaults(config)

    @classmethod
    def _animal_type_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        url_prefix = cls.get_url_prefix()

        # ajax_create
        config.add_route(f"{route_prefix}.ajax_create", f"{url_prefix}/ajax/new")
        config.add_view(
            cls,
            attr="ajax_create",
            route_name=f"{route_prefix}.ajax_create",
            permission=f"{permission_prefix}.create",
            renderer="json",
        )


class AnimalAssetView(AssetMasterView):
    """
    Master view for Animal Assets
    """

    model_class = AnimalAsset
    route_prefix = "animal_assets"
    url_prefix = "/assets/animal"

    farmos_refurl_path = "/assets/animal"
    farmos_bundle = "animal"

    labels = {
        "animal_type": "Species / Breed",
        "is_sterile": "Sterile",
    }

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "asset_name",
        "produces_eggs",
        "animal_type",
        "birthdate",
        "is_sterile",
        "sex",
        "groups",
        "owners",
        "locations",
        "archived",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model
        enum = self.app.enum

        # animal_type
        g.set_joiner("animal_type", lambda q: q.join(model.AnimalType))
        g.set_sorter("animal_type", model.AnimalType.name)
        g.set_filter("animal_type", model.AnimalType.name)
        if self.farmos_style_grid_links:
            g.set_renderer("animal_type", self.render_animal_type_for_grid)
        else:
            g.set_link("animal_type")

        # birthdate
        g.set_renderer("birthdate", "date")

        # sex
        g.set_enum("sex", enum.ANIMAL_SEX)
        g.filters["sex"].verbs = ["equal", "not_equal"]

    def render_animal_type_for_grid(self, animal, field, value):
        url = self.request.route_url("animal_types.view", uuid=animal.animal_type_uuid)
        return tags.link_to(value, url)

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        enum = self.app.enum
        animal = f.model_instance

        # animal_type
        f.fields.insert_after("asset_name", "animal_type")
        f.set_node("animal_type", AnimalTypeRef(self.request))

        # birthdate
        f.fields.insert_after("animal_type", "birthdate")
        # TODO: why must we assign the widget here?  pretty sure that
        # was not needed when we declared form_fields directly, i.e.
        # instead of adding birthdate field in this method
        f.set_widget("birthdate", WuttaDateTimeWidget(self.request))

        # produces_eggs
        f.fields.insert_after("birthdate", "produces_eggs")

        # sex
        f.fields.insert_after("produces_eggs", "sex")
        if not (self.creating or self.editing) and animal.sex is None:
            pass  # TODO: dict enum widget does not handle null values well
        else:
            # nb. ensure empty option appears like we want
            sex_enum = OrderedDict([("", "N/A")] + list(enum.ANIMAL_SEX.items()))
            f.set_node("sex", WuttaDictEnum(self.request, sex_enum))
            f.set_required("sex", False)

        # is_sterile
        f.fields.insert_after("sex", "is_sterile")


def defaults(config, **kwargs):
    base = globals()

    AnimalTypeView = kwargs.get("AnimalTypeView", base["AnimalTypeView"])
    AnimalTypeView.defaults(config)

    AnimalAssetView = kwargs.get("AnimalAssetView", base["AnimalAssetView"])
    AnimalAssetView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Common views
"""

from wuttaweb.views import common as base


class CommonView(base.CommonView):
    """
    Common views
    """

    def setup_enhance_admin_user(self, user):
        """ """
        model = self.app.model
        session = self.app.get_session(user)
        auth = self.app.get_auth_handler()

        # initialize built-in roles
        farm_manager = auth.get_role_farm_manager(session)
        farm_manager.notes = "this is meant to mirror the corresponding role in farmOS"
        farm_worker = auth.get_role_farm_worker(session)
        farm_worker.notes = "this is meant to mirror the corresponding role in farmOS"
        farm_viewer = auth.get_role_farm_viewer(session)
        farm_viewer.notes = "this is meant to mirror the corresponding role in farmOS"

        # create system user to represent farmOS
        auth.make_user(session, username="farmos", prevent_edit=True)

        site_admin = session.query(model.Role).filter_by(name="Site Admin").first()
        if site_admin:
            site_admin_perms = [
                "animal_types.create",
                "animal_types.edit",
                "animal_types.list",
                "animal_types.view",
                "animal_types.versions",
                "animal_assets.create",
                "animal_assets.edit",
                "animal_assets.list",
                "animal_assets.view",
                "animal_assets.versions",
                "assets.list",
                "asset_types.list",
                "asset_types.view",
                "asset_types.versions",
                "farmos_animal_assets.list",
                "farmos_animal_assets.view",
                "farmos_animal_types.list",
                "farmos_animal_types.view",
                "farmos_asset_types.list",
                "farmos_asset_types.view",
                "farmos_group_assets.list",
                "farmos_group_assets.view",
                "farmos_land_assets.list",
                "farmos_land_assets.view",
                "farmos_land_types.list",
                "farmos_land_types.view",
                "farmos_log_types.list",
                "farmos_log_types.view",
                "farmos_logs_activity.list",
                "farmos_logs_activity.view",
                "farmos_logs_harvest.list",
                "farmos_logs_harvest.view",
                "farmos_logs_medical.list",
                "farmos_logs_medical.view",
                "farmos_logs_observation.list",
                "farmos_logs_observation.view",
                "farmos_plant_assets.list",
                "farmos_plant_assets.view",
                "farmos_plant_types.list",
                "farmos_plant_types.view",
                "farmos_quantities_standard.list",
                "farmos_quantities_standard.view",
                "farmos_quantity_types.list",
                "farmos_quantity_types.view",
                "farmos_seasons.list",
                "farmos_seasons.view",
                "farmos_structure_assets.list",
                "farmos_structure_assets.view",
                "farmos_structure_types.list",
                "farmos_structure_types.view",
                "farmos_units.list",
                "farmos_units.view",
                "farmos_users.list",
                "farmos_users.view",
                "group_assets.create",
                "group_assets.edit",
                "group_assets.list",
                "group_assets.view",
                "group_assets.versions",
                "land_assets.create",
                "land_assets.edit",
                "land_assets.list",
                "land_assets.view",
                "land_assets.versions",
                "land_types.list",
                "land_types.view",
                "land_types.versions",
                "log_types.list",
                "log_types.view",
                "log_types.versions",
                "logs_activity.list",
                "logs_activity.view",
                "logs_activity.versions",
                "logs_harvest.list",
                "logs_harvest.view",
                "logs_harvest.versions",
                "logs_medical.list",
                "logs_medical.view",
                "logs_medical.versions",
                "logs_observation.list",
                "logs_observation.view",
                "logs_observation.versions",
                "quick.eggs",
                "plant_types.list",
                "plant_types.view",
                "plant_types.versions",
                "seasons.list",
                "seasons.view",
                "seasons.versions",
                "structure_types.list",
                "structure_types.view",
                "structure_types.versions",
                "structure_assets.create",
                "structure_assets.edit",
                "structure_assets.list",
                "structure_assets.view",
                "structure_assets.versions",
                "units.create",
                "units.edit",
                "units.list",
                "units.view",
                "units.versions",
            ]
            for perm in site_admin_perms:
                auth.grant_permission(site_admin, perm)


def defaults(config, **kwargs):
    local = globals()
    CommonView = kwargs.get("CommonView", local["CommonView"])
    base.defaults(config, **{"CommonView": CommonView})


def includeme(config):
    defaults(config)



"""use shared base for Structure Assets

Revision ID: 34ec51d80f52
Revises: d882682c82f9
Create Date: 2026-02-15 13:19:18.814523

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "34ec51d80f52"
down_revision: Union[str, None] = "d882682c82f9"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # asset_structure
    op.create_table(
        "asset_structure",
        sa.Column("structure_type_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["structure_type_uuid"],
            ["structure_type.uuid"],
            name=op.f("fk_asset_structure_structure_type_uuid_structure_type"),
        ),
        sa.ForeignKeyConstraint(
            ["uuid"], ["asset.uuid"], name=op.f("fk_asset_structure_uuid_asset")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset_structure")),
    )
    op.create_table(
        "asset_structure_version",
        sa.Column(
            "structure_type_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_asset_structure_version")
        ),
    )
    op.create_index(
        op.f("ix_asset_structure_version_end_transaction_id"),
        "asset_structure_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_structure_version_operation_type"),
        "asset_structure_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_structure_version_pk_transaction_id",
        "asset_structure_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_structure_version_pk_validity",
        "asset_structure_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_structure_version_transaction_id"),
        "asset_structure_version",
        ["transaction_id"],
        unique=False,
    )

    # structure
    op.drop_index(
        op.f("ix_structure_version_end_transaction_id"), table_name="structure_version"
    )
    op.drop_index(
        op.f("ix_structure_version_operation_type"), table_name="structure_version"
    )
    op.drop_index(
        op.f("ix_structure_version_pk_transaction_id"), table_name="structure_version"
    )
    op.drop_index(
        op.f("ix_structure_version_pk_validity"), table_name="structure_version"
    )
    op.drop_index(
        op.f("ix_structure_version_transaction_id"), table_name="structure_version"
    )
    op.drop_table("structure_version")
    op.drop_table("structure")


def downgrade() -> None:

    # structure
    op.create_table(
        "structure",
        sa.Column("uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("name", sa.VARCHAR(length=100), autoincrement=False, nullable=False),
        sa.Column("archived", sa.BOOLEAN(), autoincrement=False, nullable=False),
        sa.Column(
            "structure_type_uuid", sa.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column("is_location", sa.BOOLEAN(), autoincrement=False, nullable=False),
        sa.Column("is_fixed", sa.BOOLEAN(), autoincrement=False, nullable=False),
        sa.Column("notes", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column(
            "image_url", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.Column("farmos_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("drupal_id", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column(
            "thumbnail_url", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["structure_type_uuid"],
            ["structure_type.uuid"],
            name=op.f("fk_structure_structure_type_uuid_structure_type"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_structure")),
        sa.UniqueConstraint(
            "drupal_id",
            name=op.f("uq_structure_drupal_id"),
            postgresql_include=[],
            postgresql_nulls_not_distinct=False,
        ),
        sa.UniqueConstraint(
            "farmos_uuid",
            name=op.f("uq_structure_farmos_uuid"),
            postgresql_include=[],
            postgresql_nulls_not_distinct=False,
        ),
        sa.UniqueConstraint(
            "name",
            name=op.f("uq_structure_name"),
            postgresql_include=[],
            postgresql_nulls_not_distinct=False,
        ),
    )
    op.create_table(
        "structure_version",
        sa.Column("uuid", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("name", sa.VARCHAR(length=100), autoincrement=False, nullable=True),
        sa.Column("archived", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column("structure_type_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("is_location", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column("is_fixed", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.Column("notes", sa.TEXT(), autoincrement=False, nullable=True),
        sa.Column(
            "image_url", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.Column("farmos_uuid", sa.UUID(), autoincrement=False, nullable=True),
        sa.Column("drupal_id", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("transaction_id", sa.BIGINT(), autoincrement=False, nullable=False),
        sa.Column(
            "end_transaction_id", sa.BIGINT(), autoincrement=False, nullable=True
        ),
        sa.Column("operation_type", sa.SMALLINT(), autoincrement=False, nullable=False),
        sa.Column(
            "thumbnail_url", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_structure_version")
        ),
    )
    op.create_index(
        op.f("ix_structure_version_transaction_id"),
        "structure_version",
        ["transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_structure_version_pk_validity"),
        "structure_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_structure_version_pk_transaction_id"),
        "structure_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        op.f("ix_structure_version_operation_type"),
        "structure_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_structure_version_end_transaction_id"),
        "structure_version",
        ["end_transaction_id"],
        unique=False,
    )

    # asset_structure
    op.drop_index(
        op.f("ix_asset_structure_version_transaction_id"),
        table_name="asset_structure_version",
    )
    op.drop_index(
        "ix_asset_structure_version_pk_validity", table_name="asset_structure_version"
    )
    op.drop_index(
        "ix_asset_structure_version_pk_transaction_id",
        table_name="asset_structure_version",
    )
    op.drop_index(
        op.f("ix_asset_structure_version_operation_type"),
        table_name="asset_structure_version",
    )
    op.drop_index(
        op.f("ix_asset_structure_version_end_transaction_id"),
        table_name="asset_structure_version",
    )
    op.drop_table("asset_structure_version")
    op.drop_table("asset_structure")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base class for WuttaFarm master views
"""

from webhelpers2.html import tags

from wuttaweb.views import MasterView
from wuttaweb.util import get_form_data

from wuttafarm.web.util import use_farmos_style_grid_links, get_farmos_client_for_user


class WuttaFarmMasterView(MasterView):
    """
    Base class for WuttaFarm master views
    """

    farmos_refurl_path = None
    farmos_entity_type = None
    farmos_bundle = None

    labels = {
        "farmos_uuid": "farmOS UUID",
        "drupal_id": "Drupal ID",
        "image_url": "Image URL",
        "thumbnail_url": "Thumbnail URL",
    }

    row_labels = {
        "farmos_uuid": "farmOS UUID",
        "drupal_id": "Drupal ID",
        "image_url": "Image URL",
        "thumbnail_url": "Thumbnail URL",
    }

    def __init__(self, request, context=None):
        super().__init__(request, context=context)
        self.farmos_style_grid_links = use_farmos_style_grid_links(self.config)

    def get_farmos_url(self, obj):
        return None

    def get_template_context(self, context):

        if self.listing and self.farmos_refurl_path:
            context["farmos_refurl"] = self.app.get_farmos_url(self.farmos_refurl_path)

        return context

    def render_grid_thumbnail(self, obj, field, value):
        if obj.thumbnail_url:
            return tags.image(
                obj.thumbnail_url, f"thumbnail for {self.get_model_title()}"
            )
        return None

    def get_xref_buttons(self, obj):
        url = self.get_farmos_url(obj)
        if url:
            return [
                self.make_button(
                    "View in farmOS",
                    primary=True,
                    url=url,
                    target="_blank",
                    icon_left="external-link-alt",
                )
            ]
        return []

    def configure_form(self, form):
        """ """
        f = form
        super().configure_form(f)

        # farmos_uuid
        if self.creating:
            f.remove("farmos_uuid")
        else:
            f.set_readonly("farmos_uuid")

        # drupal_id
        if self.creating:
            f.remove("drupal_id")
        else:
            f.set_readonly("drupal_id")

    def persist(self, obj, session=None):

        # save per usual
        super().persist(obj, session)

        # maybe also sync change to farmOS
        if self.app.is_farmos_mirror():
            client = get_farmos_client_for_user(self.request)
            self.auto_sync_to_farmos(client, obj)

    def auto_sync_to_farmos(self, client, obj):
        self.app.auto_sync_to_farmos(obj, client=client, require=False)

    def get_farmos_entity_type(self):
        if self.farmos_entity_type:
            return self.farmos_entity_type
        raise NotImplementedError(
            f"must define {self.__class__.__name__}.farmos_entity_type"
        )

    def get_farmos_bundle(self):
        if self.farmos_bundle:
            return self.farmos_bundle
        raise NotImplementedError(
            f"must define {self.__class__.__name__}.farmos_bundle"
        )

    def delete_instance(self, obj):

        # save farmOS UUID if we need it
        farmos_uuid = None
        if hasattr(obj, "farmos_uuid") and self.app.is_farmos_mirror():
            farmos_uuid = obj.farmos_uuid

        # delete per usual
        super().delete_instance(obj)

        # maybe delete from farmOS also
        if farmos_uuid:
            entity_type = self.get_farmos_entity_type()
            bundle = self.get_farmos_bundle()
            client = get_farmos_client_for_user(self.request)
            client.resource.delete(entity_type, bundle, farmos_uuid)


class TaxonomyMasterView(WuttaFarmMasterView):
    """
    Base class for master views serving taxonomy terms.
    """

    farmos_entity_type = "taxonomy_term"

    grid_columns = [
        "name",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "drupal_id",
        "farmos_uuid",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # description
        f.set_widget("description", "notes")

    def get_farmos_url(self, obj):
        return self.app.get_farmos_url(f"/taxonomy/term/{obj.drupal_id}")

    def get_xref_buttons(self, term):
        buttons = super().get_xref_buttons(term)

        if term.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        f"{self.farmos_route_prefix}.view", uuid=term.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons

    def ajax_create(self):
        """
        AJAX view to create a new taxonomy term.
        """
        model = self.app.model
        session = self.Session()
        data = get_form_data(self.request)

        name = data.get("name")
        if not name:
            return {"error": "Name is required"}

        term = self.model_class(name=name)
        session.add(term)
        session.flush()

        if self.app.is_farmos_mirror():
            client = get_farmos_client_for_user(self.request)
            self.app.auto_sync_to_farmos(term, client=client)

        return {
            "uuid": term.uuid.hex,
            "name": term.name,
            "farmos_uuid": term.farmos_uuid.hex,
            "drupal_id": term.drupal_id,
        }

    @classmethod
    def defaults(cls, config):
        """ """
        cls._defaults(config)
        cls._taxonomy_defaults(config)

    @classmethod
    def _taxonomy_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        url_prefix = cls.get_url_prefix()

        # ajax_create
        config.add_route(f"{route_prefix}.ajax_create", f"{url_prefix}/ajax/new")
        config.add_view(
            cls,
            attr="ajax_create",
            route_name=f"{route_prefix}.ajax_create",
            permission=f"{permission_prefix}.create",
            renderer="json",
        )



"""remove unique for animal_type.name

Revision ID: 45c7718d2ed2
Revises: 5b6c87d8cddf
Create Date: 2026-02-27 16:53:59.310342

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "45c7718d2ed2"
down_revision: Union[str, None] = "5b6c87d8cddf"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # animal_type
    op.drop_constraint(op.f("uq_animal_type_name"), "animal_type", type_="unique")


def downgrade() -> None:

    # animal_type
    op.create_unique_constraint(
        op.f("uq_animal_type_name"),
        "animal_type",
        ["name"],
        postgresql_nulls_not_distinct=False,
    )



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Farm Plants
"""

import datetime

import colander

from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos.master import TaxonomyMasterView
from wuttafarm.web.views.farmos import FarmOSMasterView
from wuttafarm.web.forms.schema import (
    UsersType,
    StructureType,
    FarmOSPlantTypes,
    FarmOSRefs,
)
from wuttafarm.web.forms.widgets import ImageWidget


class PlantTypeView(TaxonomyMasterView):
    """
    Master view for Plant Types in farmOS.
    """

    model_name = "farmos_plant_type"
    model_title = "farmOS Plant Type"
    model_title_plural = "farmOS Plant Types"

    route_prefix = "farmos_plant_types"
    url_prefix = "/farmOS/plant-types"

    farmos_taxonomy_type = "plant_type"
    farmos_refurl_path = "/admin/structure/taxonomy/manage/plant_type/overview"

    def get_xref_buttons(self, plant_type):
        buttons = super().get_xref_buttons(plant_type)
        model = self.app.model
        session = self.Session()

        if wf_plant_type := (
            session.query(model.PlantType)
            .filter(model.PlantType.farmos_uuid == plant_type["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "plant_types.view", uuid=wf_plant_type.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class SeasonView(TaxonomyMasterView):
    """
    Master view for Seasons in farmOS.
    """

    model_name = "farmos_season"
    model_title = "farmOS Season"
    model_title_plural = "farmOS Seasons"

    route_prefix = "farmos_seasons"
    url_prefix = "/farmOS/seasons"

    farmos_taxonomy_type = "season"
    farmos_refurl_path = "/admin/structure/taxonomy/manage/season/overview"

    def get_xref_buttons(self, season):
        buttons = super().get_xref_buttons(season)
        model = self.app.model
        session = self.Session()

        if wf_season := (
            session.query(model.Season)
            .filter(model.Season.farmos_uuid == season["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url("seasons.view", uuid=wf_season.uuid),
                    icon_left="eye",
                )
            )

        return buttons


class PlantAssetView(FarmOSMasterView):
    """
    Master view for farmOS Plant Assets
    """

    model_name = "farmos_plant_assets"
    model_title = "farmOS Plant Asset"
    model_title_plural = "farmOS Plant Assets"

    route_prefix = "farmos_plant_assets"
    url_prefix = "/farmOS/assets/plant"

    farmos_refurl_path = "/assets/plant"

    labels = {
        "seasons": "Season",
    }

    grid_columns = [
        "name",
        "archived",
    ]

    sort_defaults = "name"

    form_fields = [
        "name",
        "plant_types",
        "seasons",
        "archived",
        "owners",
        "location",
        "notes",
        "raw_image_url",
        "large_image_url",
        "thumbnail_image_url",
        "image",
    ]

    def get_grid_data(self, columns=None, session=None):
        result = self.farmos_client.asset.get("plant")
        return [self.normalize_plant(a) for a in result["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")
        g.set_searchable("name")

        # archived
        g.set_renderer("archived", "boolean")

    def get_instance(self):

        plant = self.farmos_client.resource.get_id(
            "asset", "plant", self.request.matchdict["uuid"]
        )
        self.raw_json = plant

        # instance data
        data = self.normalize_plant(plant["data"])

        if relationships := plant["data"].get("relationships"):

            # add plant types
            if plant_type := relationships.get("plant_type"):
                if plant_type["data"]:
                    data["plant_types"] = []
                    for plant_type in plant_type["data"]:
                        plant_type = self.farmos_client.resource.get_id(
                            "taxonomy_term", "plant_type", plant_type["id"]
                        )
                        data["plant_types"].append(
                            {
                                "uuid": plant_type["data"]["id"],
                                "name": plant_type["data"]["attributes"]["name"],
                            }
                        )

            # add seasons
            if seasons := relationships.get("season"):
                if seasons["data"]:
                    data["seasons"] = []
                    for season in seasons["data"]:
                        season = self.farmos_client.resource.get_id(
                            "taxonomy_term", "season", season["id"]
                        )
                        data["seasons"].append(
                            {
                                "uuid": season["data"]["id"],
                                "name": season["data"]["attributes"]["name"],
                            }
                        )

            # add location
            if location := relationships.get("location"):
                if location["data"]:
                    location = self.farmos_client.resource.get_id(
                        "asset", "structure", location["data"][0]["id"]
                    )
                    data["location"] = {
                        "uuid": location["data"]["id"],
                        "name": location["data"]["attributes"]["name"],
                    }

            # add owners
            if owner := relationships.get("owner"):
                data["owners"] = []
                for owner_data in owner["data"]:
                    owner = self.farmos_client.resource.get_id(
                        "user", "user", owner_data["id"]
                    )
                    data["owners"].append(
                        {
                            "uuid": owner["data"]["id"],
                            "display_name": owner["data"]["attributes"]["display_name"],
                        }
                    )

            # add image urls
            if image := relationships.get("image"):
                if image["data"]:
                    image = self.farmos_client.resource.get_id(
                        "file", "file", image["data"][0]["id"]
                    )
                    data["raw_image_url"] = self.app.get_farmos_url(
                        image["data"]["attributes"]["uri"]["url"]
                    )
                    # nb. other styles available:  medium, wide
                    data["large_image_url"] = image["data"]["attributes"][
                        "image_style_uri"
                    ]["large"]
                    data["thumbnail_image_url"] = image["data"]["attributes"][
                        "image_style_uri"
                    ]["thumbnail"]

        return data

    def get_instance_title(self, plant):
        return plant["name"]

    def normalize_plant(self, plant):
        normal = self.normal.normalize_farmos_asset(plant)
        return {
            "uuid": normal["uuid"],
            "drupal_id": normal["drupal_id"],
            "name": normal["asset_name"],
            "location": colander.null,  # TODO
            "archived": normal["archived"],
            "notes": normal["notes"] or colander.null,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        plant = f.model_instance

        # plant_types
        f.set_node("plant_types", FarmOSPlantTypes(self.request))

        # seasons
        f.set_node("seasons", FarmOSRefs(self.request, "farmos_seasons"))

        # location
        f.set_node("location", StructureType(self.request))

        # owners
        f.set_node("owners", UsersType(self.request))

        # notes
        f.set_widget("notes", "notes")

        # archived
        f.set_node("archived", colander.Boolean())

        # image
        if url := plant.get("large_image_url"):
            f.set_widget("image", ImageWidget("plant image"))
            f.set_default("image", url)

    def get_xref_buttons(self, plant):
        model = self.app.model
        session = self.Session()

        buttons = [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(f"/asset/{plant['drupal_id']}"),
                target="_blank",
                icon_left="external-link-alt",
            ),
        ]

        if wf_plant := (
            session.query(model.Asset)
            .filter(model.Asset.farmos_uuid == plant["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url("plant_assets.view", uuid=wf_plant.uuid),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    PlantTypeView = kwargs.get("PlantTypeView", base["PlantTypeView"])
    PlantTypeView.defaults(config)

    SeasonView = kwargs.get("SeasonView", base["SeasonView"])
    SeasonView.defaults(config)

    PlantAssetView = kwargs.get("PlantAssetView", base["PlantAssetView"])
    PlantAssetView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Observation Logs
"""

from wuttafarm.web.views.logs import LogMasterView
from wuttafarm.db.model import ObservationLog


class ObservationLogView(LogMasterView):
    """
    Master view for Observation Logs
    """

    model_class = ObservationLog
    route_prefix = "logs_observation"
    url_prefix = "/logs/observation"

    farmos_bundle = "observation"
    farmos_refurl_path = "/logs/observation"

    grid_columns = [
        "status",
        "drupal_id",
        "timestamp",
        "message",
        "assets",
        "locations",
        "groups",
        "is_group_assignment",
        "owners",
    ]


def defaults(config, **kwargs):
    base = globals()

    ObservationLogView = kwargs.get("ObservationLogView", base["ObservationLogView"])
    ObservationLogView.defaults(config)


def includeme(config):
    defaults(config)



"""add generic log base

Revision ID: dd6351e69233
Revises: b8cd4a8f981f
Create Date: 2026-02-18 12:09:05.200134

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "dd6351e69233"
down_revision: Union[str, None] = "b8cd4a8f981f"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # log
    op.create_table(
        "log",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("log_type", sa.String(length=100), nullable=False),
        sa.Column("message", sa.String(length=255), nullable=False),
        sa.Column("timestamp", sa.DateTime(), nullable=False),
        sa.Column("status", sa.String(length=20), nullable=False),
        sa.Column("notes", sa.Text(), nullable=True),
        sa.Column("farmos_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.Column("drupal_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["log_type"], ["log_type.drupal_id"], name=op.f("fk_log_log_type_log_type")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_log")),
        sa.UniqueConstraint("drupal_id", name=op.f("uq_log_drupal_id")),
        sa.UniqueConstraint("farmos_uuid", name=op.f("uq_log_farmos_uuid")),
    )
    op.create_table(
        "log_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "log_type", sa.String(length=100), autoincrement=False, nullable=True
        ),
        sa.Column("message", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column("timestamp", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column("status", sa.String(length=20), autoincrement=False, nullable=True),
        sa.Column("notes", sa.Text(), autoincrement=False, nullable=True),
        sa.Column(
            "farmos_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("drupal_id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("uuid", "transaction_id", name=op.f("pk_log_version")),
    )
    op.create_index(
        op.f("ix_log_version_end_transaction_id"),
        "log_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_version_operation_type"),
        "log_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_log_version_pk_transaction_id",
        "log_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_log_version_pk_validity",
        "log_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_version_transaction_id"),
        "log_version",
        ["transaction_id"],
        unique=False,
    )

    # log_activity
    op.drop_column("log_activity_version", "status")
    op.drop_column("log_activity_version", "farmos_uuid")
    op.drop_column("log_activity_version", "timestamp")
    op.drop_column("log_activity_version", "message")
    op.drop_column("log_activity_version", "drupal_id")
    op.drop_column("log_activity_version", "notes")
    op.drop_constraint(
        op.f("uq_log_activity_drupal_id"), "log_activity", type_="unique"
    )
    op.drop_constraint(
        op.f("uq_log_activity_farmos_uuid"), "log_activity", type_="unique"
    )
    op.create_foreign_key(
        op.f("fk_log_activity_uuid_log"), "log_activity", "log", ["uuid"], ["uuid"]
    )
    op.drop_column("log_activity", "status")
    op.drop_column("log_activity", "farmos_uuid")
    op.drop_column("log_activity", "timestamp")
    op.drop_column("log_activity", "message")
    op.drop_column("log_activity", "drupal_id")
    op.drop_column("log_activity", "notes")


def downgrade() -> None:

    # log_activity
    op.add_column(
        "log_activity",
        sa.Column("notes", sa.TEXT(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "log_activity",
        sa.Column("drupal_id", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "log_activity",
        sa.Column(
            "message", sa.VARCHAR(length=255), autoincrement=False, nullable=False
        ),
    )
    op.add_column(
        "log_activity",
        sa.Column(
            "timestamp", postgresql.TIMESTAMP(), autoincrement=False, nullable=False
        ),
    )
    op.add_column(
        "log_activity",
        sa.Column("farmos_uuid", sa.UUID(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "log_activity",
        sa.Column("status", sa.VARCHAR(length=20), autoincrement=False, nullable=False),
    )
    op.drop_constraint(
        op.f("fk_log_activity_uuid_log"), "log_activity", type_="foreignkey"
    )
    op.create_unique_constraint(
        op.f("uq_log_activity_farmos_uuid"),
        "log_activity",
        ["farmos_uuid"],
        postgresql_nulls_not_distinct=False,
    )
    op.create_unique_constraint(
        op.f("uq_log_activity_drupal_id"),
        "log_activity",
        ["drupal_id"],
        postgresql_nulls_not_distinct=False,
    )
    op.add_column(
        "log_activity_version",
        sa.Column("notes", sa.TEXT(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "log_activity_version",
        sa.Column("drupal_id", sa.INTEGER(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "log_activity_version",
        sa.Column(
            "message", sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "log_activity_version",
        sa.Column(
            "timestamp", postgresql.TIMESTAMP(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "log_activity_version",
        sa.Column("farmos_uuid", sa.UUID(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "log_activity_version",
        sa.Column("status", sa.VARCHAR(length=20), autoincrement=False, nullable=True),
    )

    # log
    op.drop_index(op.f("ix_log_version_transaction_id"), table_name="log_version")
    op.drop_index("ix_log_version_pk_validity", table_name="log_version")
    op.drop_index("ix_log_version_pk_transaction_id", table_name="log_version")
    op.drop_index(op.f("ix_log_version_operation_type"), table_name="log_version")
    op.drop_index(op.f("ix_log_version_end_transaction_id"), table_name="log_version")
    op.drop_table("log_version")
    op.drop_table("log")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data import for farmOS -> WuttaFarm
"""

import datetime
import logging
from uuid import UUID

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from wuttasync.importing import ImportHandler, ToWuttaHandler, Importer, ToWutta

from wuttafarm.db import model


log = logging.getLogger(__name__)


class FromFarmOSHandler(ImportHandler):
    """
    Base class for import handler using farmOS API as data source.
    """

    source_key = "farmos"
    generic_source_title = "farmOS"

    def begin_source_transaction(self, client=None):
        """
        Establish the farmOS API client.
        """
        if client:
            self.farmos_client = client
        else:
            token = self.get_farmos_oauth2_token()
            self.farmos_client = self.app.get_farmos_client(token=token)
        self.farmos_4x = self.app.is_farmos_4x(self.farmos_client)
        self.normal = self.app.get_normalizer(self.farmos_client)

    def get_farmos_oauth2_token(self):

        client_id = self.config.get(
            "farmos.oauth2.importing.client_id", default="wuttafarm"
        )
        client_secret = self.config.require("farmos.oauth2.importing.client_secret")
        scope = self.config.get("farmos.oauth2.importing.scope", default="farm_manager")

        client = BackendApplicationClient(client_id=client_id)
        oauth = OAuth2Session(client=client)

        return oauth.fetch_token(
            token_url=self.app.get_farmos_url("/oauth/token"),
            include_client_id=True,
            client_secret=client_secret,
            scope=scope,
        )

    def get_importer_kwargs(self, key, **kwargs):
        kwargs = super().get_importer_kwargs(key, **kwargs)
        kwargs["farmos_client"] = self.farmos_client
        kwargs["farmos_4x"] = self.farmos_4x
        kwargs["normal"] = self.normal
        return kwargs


class ToWuttaFarmHandler(ToWuttaHandler):
    """
    Base class for import handler targeting WuttaFarm
    """

    target_key = "wuttafarm"


class FromFarmOSToWuttaFarm(FromFarmOSHandler, ToWuttaFarmHandler):
    """
    Handler for farmOS → WuttaFarm import.
    """

    def define_importers(self):
        """ """
        importers = super().define_importers()
        importers["User"] = UserImporter
        importers["AssetType"] = AssetTypeImporter
        importers["LandType"] = LandTypeImporter
        importers["LandAsset"] = LandAssetImporter
        importers["StructureType"] = StructureTypeImporter
        importers["StructureAsset"] = StructureAssetImporter
        importers["WaterAsset"] = WaterAssetImporter
        importers["EquipmentType"] = EquipmentTypeImporter
        importers["EquipmentAsset"] = EquipmentAssetImporter
        importers["AnimalType"] = AnimalTypeImporter
        importers["AnimalAsset"] = AnimalAssetImporter
        importers["GroupAsset"] = GroupAssetImporter
        importers["PlantType"] = PlantTypeImporter
        importers["Season"] = SeasonImporter
        importers["PlantAsset"] = PlantAssetImporter
        importers["Measure"] = MeasureImporter
        importers["Unit"] = UnitImporter
        importers["MaterialType"] = MaterialTypeImporter
        importers["QuantityType"] = QuantityTypeImporter
        importers["StandardQuantity"] = StandardQuantityImporter
        importers["MaterialQuantity"] = MaterialQuantityImporter
        importers["LogType"] = LogTypeImporter
        importers["ActivityLog"] = ActivityLogImporter
        importers["HarvestLog"] = HarvestLogImporter
        importers["MedicalLog"] = MedicalLogImporter
        importers["ObservationLog"] = ObservationLogImporter
        importers["SeedingLog"] = SeedingLogImporter
        return importers


class FromFarmOS(Importer):
    """
    Base class for importers using farmOS API as data source.
    """

    key = "farmos_uuid"

    def get_supported_fields(self):
        """
        Auto-remove the ``uuid`` field, since we use ``farmos_uuid``
        instead for the importer key.
        """
        fields = list(super().get_supported_fields())
        if "uuid" in fields:
            fields.remove("uuid")
        return fields

    def normalize_datetime(self, dt):
        """
        Convert a farmOS datetime value to naive UTC used by
        WuttaFarm.

        :param dt: Date/time string value "as-is" from the farmOS API.

        :returns: Equivalent naive UTC ``datetime``
        """
        if not dt:
            return None
        dt = datetime.datetime.fromisoformat(dt)
        return self.app.make_utc(dt)


class AssetImporterBase(FromFarmOS, ToWutta):
    """
    Base class for farmOS API → WuttaFarm asset importers
    """

    def get_farmos_asset_type(self):
        return self.model_class.__wutta_hint__["farmos_asset_type"]

    def get_simple_fields(self):
        """ """
        fields = list(super().get_simple_fields())
        # nb. must explicitly declare proxy fields
        fields.extend(
            [
                "farmos_uuid",
                "drupal_id",
                "asset_type",
                "asset_name",
                "is_location",
                "is_fixed",
                "notes",
                "archived",
                "image_url",
                "thumbnail_url",
            ]
        )
        return fields

    def get_supported_fields(self):
        """ """
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "parents",
                "owners",
            ]
        )
        return fields

    def get_source_objects(self):
        """ """
        asset_type = self.get_farmos_asset_type()
        return list(
            self.farmos_client.asset.iterate(asset_type, params={"include": "image"})
        )

    def normalize_source_data(self, **kwargs):
        """ """
        data = super().normalize_source_data(**kwargs)

        if "parents" in self.fields:
            # nb. make sure parent-less (root) assets come first, so they
            # exist when child assets need to reference them
            data.sort(key=lambda l: len(l["parents"]))

        return data

    def normalize_source_object(self, asset):
        """ """
        data = self.normal.normalize_farmos_asset(asset)

        data["farmos_uuid"] = UUID(data.pop("uuid"))
        data["asset_type"] = self.get_asset_type(asset)

        if "image_url" in self.fields or "thumbnail_url" in self.fields:
            data["image_url"] = None
            data["thumbnail_url"] = None
            if relationships := asset.get("relationships"):

                if image := relationships.get("image"):
                    if image["data"]:
                        image = self.farmos_client.resource.get_id(
                            "file", "file", image["data"][0]["id"]
                        )
                        if image_style := image["data"]["attributes"].get(
                            "image_style_uri"
                        ):
                            data["image_url"] = image_style["large"]
                            data["thumbnail_url"] = image_style["thumbnail"]

        if "parents" in self.fields:
            data["parents"] = []
            for parent in asset["relationships"]["parent"]["data"]:
                data["parents"].append(
                    (self.get_asset_type(parent), UUID(parent["id"]))
                )

        if "owners" in self.fields:
            data["owners"] = [UUID(uuid) for uuid in data["owner_uuids"]]

        return data

    def get_asset_type(self, asset):
        return asset["type"].split("--")[1]

    def normalize_target_object(self, asset):
        data = super().normalize_target_object(asset)

        if "parents" in self.fields:
            data["parents"] = [(p.asset_type, p.farmos_uuid) for p in asset.parents]

        if "owners" in self.fields:
            data["owners"] = [user.farmos_uuid for user in asset.owners]

        return data

    def update_target_object(self, asset, source_data, target_data=None):
        model = self.app.model
        asset = super().update_target_object(asset, source_data, target_data)

        if "parents" in self.fields:
            if not target_data or target_data["parents"] != source_data["parents"]:

                for key in source_data["parents"]:
                    asset_type, farmos_uuid = key
                    if not target_data or key not in target_data["parents"]:
                        self.target_session.flush()
                        parent = (
                            self.target_session.query(model.Asset)
                            .filter(model.Asset.asset_type == asset_type)
                            .filter(model.Asset.farmos_uuid == farmos_uuid)
                            .one()
                        )
                        asset.asset._parents.append(model.AssetParent(parent=parent))

                if target_data:
                    for key in target_data["parents"]:
                        asset_type, farmos_uuid = key
                        if key not in source_data["parents"]:
                            parent = (
                                self.target_session.query(model.Asset)
                                .filter(model.Asset.asset_type == asset_type)
                                .filter(model.Asset.farmos_uuid == farmos_uuid)
                                .one()
                            )
                            parent = (
                                self.target_session.query(model.AssetParent)
                                .filter(model.AssetParent.asset == asset)
                                .filter(model.AssetParent.parent == parent)
                                .one()
                            )
                            self.target_session.delete(parent)

        if "owners" in self.fields:
            if not target_data or target_data["owners"] != source_data["owners"]:

                for farmos_uuid in source_data["owners"]:
                    if not target_data or farmos_uuid not in target_data["owners"]:
                        user = (
                            self.target_session.query(model.User)
                            .join(model.WuttaFarmUser)
                            .filter(model.WuttaFarmUser.farmos_uuid == farmos_uuid)
                            .one()
                        )
                        asset.owners.append(user)

                if target_data:
                    for farmos_uuid in target_data["owners"]:
                        if farmos_uuid not in source_data["owners"]:
                            user = (
                                self.target_session.query(model.User)
                                .join(model.WuttaFarmUser)
                                .filter(model.WuttaFarmUser.farmos_uuid == farmos_uuid)
                                .one()
                            )
                            asset.owners.remove(user)

        return asset


class AnimalAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Animals
    """

    model_class = model.AnimalAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "animal_type_uuid",
                "sex",
                "is_sterile",
                "produces_eggs",
                "birthdate",
            ]
        )
        return fields

    def setup(self):
        super().setup()
        model = self.app.model

        self.animal_types_by_farmos_uuid = {}
        for animal_type in self.target_session.query(model.AnimalType):
            if animal_type.farmos_uuid:
                self.animal_types_by_farmos_uuid[animal_type.farmos_uuid] = animal_type

    def normalize_source_object(self, animal):
        """ """
        animal_type_uuid = None
        if relationships := animal.get("relationships"):

            if animal_type := relationships.get("animal_type"):
                if animal_type["data"]:
                    if wf_animal_type := self.animal_types_by_farmos_uuid.get(
                        UUID(animal_type["data"]["id"])
                    ):
                        animal_type_uuid = wf_animal_type.uuid
                    else:
                        log.warning(
                            "animal type not found: %s", animal_type["data"]["id"]
                        )

        if not animal_type_uuid:
            log.warning("missing/invalid animal_type for farmOS Animal: %s", animal)
            return None

        birthdate = animal["attributes"]["birthdate"]
        if birthdate:
            birthdate = datetime.datetime.fromisoformat(birthdate)
            birthdate = self.app.localtime(birthdate)
            birthdate = self.app.make_utc(birthdate)

        if self.farmos_4x:
            sterile = animal["attributes"]["is_sterile"]
        else:
            sterile = animal["attributes"]["is_castrated"]

        data = super().normalize_source_object(animal)
        data.update(
            {
                "animal_type_uuid": animal_type_uuid,
                "sex": animal["attributes"]["sex"],
                "is_sterile": sterile,
                "produces_eggs": animal["attributes"]["produces_eggs"],
                "birthdate": birthdate,
            }
        )
        return data


class TaxonomyImporterBase(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for taxonomy terms
    """

    taxonomy_type = None

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
    ]

    def get_source_objects(self):
        """ """
        return list(
            self.farmos_client.resource.iterate("taxonomy_term", self.taxonomy_type)
        )

    def normalize_source_object(self, term):
        """ """
        if description := term["attributes"]["description"]:
            description = description["value"]

        return {
            "farmos_uuid": UUID(term["id"]),
            "drupal_id": term["attributes"]["drupal_internal__tid"],
            "name": term["attributes"]["name"],
            "description": description,
        }


class AnimalTypeImporter(TaxonomyImporterBase):
    """
    farmOS API → WuttaFarm importer for Animal Types
    """

    model_class = model.AnimalType
    taxonomy_type = "animal_type"


class MaterialTypeImporter(TaxonomyImporterBase):
    """
    farmOS API → WuttaFarm importer for Material Types
    """

    model_class = model.MaterialType
    taxonomy_type = "material_type"


class EquipmentTypeImporter(TaxonomyImporterBase):
    """
    farmOS API → WuttaFarm importer for Equipment Types
    """

    model_class = model.EquipmentType
    taxonomy_type = "equipment_type"


class AssetTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Asset Types
    """

    model_class = model.AssetType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
    ]

    def get_source_objects(self):
        """ """
        return list(self.farmos_client.resource.iterate("asset_type"))

    def normalize_source_object(self, asset_type):
        """ """
        return {
            "farmos_uuid": UUID(asset_type["id"]),
            "drupal_id": asset_type["attributes"]["drupal_internal__id"],
            "name": asset_type["attributes"]["label"],
            "description": asset_type["attributes"]["description"],
        }


class EquipmentAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Equipment Assets
    """

    model_class = model.EquipmentAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "equipment_types",
            ]
        )
        return fields

    def setup(self):
        super().setup()
        model = self.app.model

        self.equipment_types_by_farmos_uuid = {}
        for equipment_type in self.target_session.query(model.EquipmentType):
            if equipment_type.farmos_uuid:
                self.equipment_types_by_farmos_uuid[equipment_type.farmos_uuid] = (
                    equipment_type
                )

    def normalize_source_object(self, equipment):
        """ """
        data = super().normalize_source_object(equipment)

        equipment_types = []
        if relationships := equipment.get("relationships"):

            if equipment_type := relationships.get("equipment_type"):
                equipment_types = []
                for equipment_type in equipment_type["data"]:
                    if wf_equipment_type := self.equipment_types_by_farmos_uuid.get(
                        UUID(equipment_type["id"])
                    ):
                        equipment_types.append(wf_equipment_type.uuid)
                    else:
                        log.warning(
                            "equipment type not found: %s", equipment_type["id"]
                        )

        data.update(
            {
                "manufacturer": equipment["attributes"]["manufacturer"],
                "model": equipment["attributes"]["model"],
                "serial_number": equipment["attributes"]["serial_number"],
                "equipment_types": set(equipment_types),
            }
        )
        return data

    def normalize_target_object(self, equipment):
        data = super().normalize_target_object(equipment)

        if "equipment_types" in self.fields:
            data["equipment_types"] = set(
                [etype.uuid for etype in equipment.equipment_types]
            )

        return data

    def update_target_object(self, equipment, source_data, target_data=None):
        model = self.app.model
        equipment = super().update_target_object(equipment, source_data, target_data)

        if "equipment_types" in self.fields:
            if (
                not target_data
                or target_data["equipment_types"] != source_data["equipment_types"]
            ):

                for uuid in source_data["equipment_types"]:
                    if not target_data or uuid not in target_data["equipment_types"]:
                        self.target_session.flush()
                        equipment._equipment_types.append(
                            model.EquipmentAssetEquipmentType(equipment_type_uuid=uuid)
                        )

                if target_data:
                    for uuid in target_data["equipment_types"]:
                        if uuid not in source_data["equipment_types"]:
                            equipment_type = (
                                self.target_session.query(
                                    model.EquipmentAssetEquipmentType
                                )
                                .filter(
                                    model.EquipmentAssetEquipmentType.equipment_asset
                                    == equipment
                                )
                                .filter(
                                    model.EquipmentAssetEquipmentType.equipment_type_uuid
                                    == uuid
                                )
                                .one()
                            )
                            self.target_session.delete(equipment_type)

        return equipment


class GroupAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Group Assets
    """

    model_class = model.GroupAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "produces_eggs",
            ]
        )
        return fields

    def normalize_source_object(self, group):
        """ """
        data = super().normalize_source_object(group)
        data.update(
            {
                "produces_eggs": group["attributes"]["produces_eggs"],
            }
        )
        return data


class LandAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Land Assets
    """

    model_class = model.LandAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "land_type_uuid",
            ]
        )
        return fields

    def setup(self):
        """ """
        super().setup()
        model = self.app.model

        self.land_types_by_id = {}
        for land_type in self.target_session.query(model.LandType):
            self.land_types_by_id[land_type.drupal_id] = land_type

    def normalize_source_object(self, land):
        """ """
        land_type_id = land["attributes"]["land_type"]
        land_type = self.land_types_by_id.get(land_type_id)
        if not land_type:
            log.warning(
                "invalid land_type '%s' for farmOS Land Asset: %s", land_type_id, land
            )
            return None

        data = super().normalize_source_object(land)
        data.update(
            {
                "land_type_uuid": land_type.uuid,
            }
        )
        return data


class LandTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Land Types
    """

    model_class = model.LandType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
    ]

    def get_source_objects(self):
        """ """
        return list(self.farmos_client.resource.iterate("land_type"))

    def normalize_source_object(self, land_type):
        """ """
        return {
            "farmos_uuid": UUID(land_type["id"]),
            "drupal_id": land_type["attributes"]["drupal_internal__id"],
            "name": land_type["attributes"]["label"],
        }


class PlantTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Plant Types
    """

    model_class = model.PlantType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
    ]

    def get_source_objects(self):
        """ """
        return list(self.farmos_client.resource.iterate("taxonomy_term", "plant_type"))

    def normalize_source_object(self, plant_type):
        """ """
        return {
            "farmos_uuid": UUID(plant_type["id"]),
            "drupal_id": plant_type["attributes"]["drupal_internal__tid"],
            "name": plant_type["attributes"]["name"],
            "description": plant_type["attributes"]["description"],
        }


class SeasonImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Seasons
    """

    model_class = model.Season

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
    ]

    def get_source_objects(self):
        """ """
        return list(self.farmos_client.resource.iterate("taxonomy_term", "season"))

    def normalize_source_object(self, season):
        """ """
        return {
            "farmos_uuid": UUID(season["id"]),
            "drupal_id": season["attributes"]["drupal_internal__tid"],
            "name": season["attributes"]["name"],
            "description": season["attributes"]["description"],
        }


class PlantAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Plant Assets
    """

    model_class = model.PlantAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "plant_types",
                "seasons",
            ]
        )
        return fields

    def setup(self):
        super().setup()
        model = self.app.model

        self.plant_types_by_farmos_uuid = {}
        for plant_type in self.target_session.query(model.PlantType):
            if plant_type.farmos_uuid:
                self.plant_types_by_farmos_uuid[plant_type.farmos_uuid] = plant_type

        self.seasons_by_farmos_uuid = {}
        for season in self.target_session.query(model.Season):
            if season.farmos_uuid:
                self.seasons_by_farmos_uuid[season.farmos_uuid] = season

    def normalize_source_object(self, plant):
        """ """
        data = super().normalize_source_object(plant)

        plant_types = []
        seasons = []
        if relationships := plant.get("relationships"):

            if plant_type := relationships.get("plant_type"):
                plant_types = []
                for plant_type in plant_type["data"]:
                    if wf_plant_type := self.plant_types_by_farmos_uuid.get(
                        UUID(plant_type["id"])
                    ):
                        plant_types.append(wf_plant_type.uuid)
                    else:
                        log.warning("plant type not found: %s", plant_type["id"])

            if season := relationships.get("season"):
                seasons = []
                for season in season["data"]:
                    if wf_season := self.seasons_by_farmos_uuid.get(UUID(season["id"])):
                        seasons.append(wf_season.uuid)
                    else:
                        log.warning("season not found: %s", season["id"])

        data.update(
            {
                "plant_types": set(plant_types),
                "seasons": set(seasons),
            }
        )
        return data

    def normalize_target_object(self, plant):
        data = super().normalize_target_object(plant)

        if "plant_types" in self.fields:
            data["plant_types"] = set([pt.uuid for pt in plant.plant_types])

        if "seasons" in self.fields:
            data["seasons"] = set([s.uuid for s in plant.seasons])

        return data

    def update_target_object(self, plant, source_data, target_data=None):
        model = self.app.model
        plant = super().update_target_object(plant, source_data, target_data)

        if "plant_types" in self.fields:
            if (
                not target_data
                or target_data["plant_types"] != source_data["plant_types"]
            ):

                for uuid in source_data["plant_types"]:
                    if not target_data or uuid not in target_data["plant_types"]:
                        self.target_session.flush()
                        plant._plant_types.append(
                            model.PlantAssetPlantType(plant_type_uuid=uuid)
                        )

                if target_data:
                    for uuid in target_data["plant_types"]:
                        if uuid not in source_data["plant_types"]:
                            plant_type = (
                                self.target_session.query(model.PlantAssetPlantType)
                                .filter(model.PlantAssetPlantType.plant_asset == plant)
                                .filter(
                                    model.PlantAssetPlantType.plant_type_uuid == uuid
                                )
                                .one()
                            )
                            self.target_session.delete(plant_type)

        if "seasons" in self.fields:
            if not target_data or target_data["seasons"] != source_data["seasons"]:

                for uuid in source_data["seasons"]:
                    if not target_data or uuid not in target_data["seasons"]:
                        self.target_session.flush()
                        plant._seasons.append(model.PlantAssetSeason(season_uuid=uuid))

                if target_data:
                    for uuid in target_data["seasons"]:
                        if uuid not in source_data["seasons"]:
                            season = (
                                self.target_session.query(model.PlantAssetSeason)
                                .filter(model.PlantAssetSeason.plant_asset == plant)
                                .filter(model.PlantAssetSeason.season_uuid == uuid)
                                .one()
                            )
                            self.target_session.delete(season)

        return plant


class StructureAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Structure Assets
    """

    model_class = model.StructureAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "structure_type_uuid",
            ]
        )
        return fields

    def setup(self):
        super().setup()
        model = self.app.model

        self.structure_types_by_id = {}
        for structure_type in self.target_session.query(model.StructureType):
            self.structure_types_by_id[structure_type.drupal_id] = structure_type

    def normalize_source_object(self, structure):
        """ """
        structure_type_id = structure["attributes"]["structure_type"]
        structure_type = self.structure_types_by_id.get(structure_type_id)
        if not structure_type:
            log.warning(
                "invalid structure_type '%s' for farmOS Structure Asset: %s",
                structure_type_id,
                structure,
            )
            return None

        data = super().normalize_source_object(structure)
        data.update(
            {
                "structure_type_uuid": structure_type.uuid,
            }
        )
        return data


class StructureTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Structure Types
    """

    model_class = model.StructureType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
    ]

    def get_source_objects(self):
        """ """
        return list(self.farmos_client.resource.iterate("structure_type"))

    def normalize_source_object(self, structure_type):
        """ """
        return {
            "farmos_uuid": UUID(structure_type["id"]),
            "drupal_id": structure_type["attributes"]["drupal_internal__id"],
            "name": structure_type["attributes"]["label"],
        }


class WaterAssetImporter(AssetImporterBase):
    """
    farmOS API → WuttaFarm importer for Water Assets
    """

    model_class = model.WaterAsset


class UserImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Users
    """

    model_class = model.User

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "username",
    ]

    def get_simple_fields(self):
        """ """
        fields = list(super().get_simple_fields())
        # nb. must explicitly declare extension fields
        fields.extend(
            [
                "farmos_uuid",
                "drupal_id",
            ]
        )
        return fields

    def get_source_objects(self):
        """ """
        return list(self.farmos_client.resource.iterate("user"))

    def normalize_source_object(self, user):
        """ """

        # nb. skip Anonymous user which does not have drupal id
        drupal_id = user["attributes"].get("drupal_internal__uid")
        if not drupal_id:
            return None

        return {
            "farmos_uuid": UUID(user["id"]),
            "drupal_id": drupal_id,
            "username": user["attributes"]["name"],
        }

    def can_delete_object(self, user, data=None):
        """
        Prevent delete for users which do not exist in farmOS.
        """
        if not user.farmos_uuid:
            return False
        return True


##############################
# log importers
##############################


class MeasureImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Measures
    """

    model_class = model.Measure

    key = "drupal_id"

    supported_fields = [
        "drupal_id",
        "ordinal",
        "name",
    ]

    def get_source_objects(self):
        """ """
        response = self.farmos_client.session.get(
            self.app.get_farmos_url("/api/quantity/standard/resource/schema")
        )
        response.raise_for_status()
        data = response.json()
        self.ordinal = 0
        return data["definitions"]["attributes"]["properties"]["measure"]["oneOf"]

    def normalize_source_object(self, measure):
        """ """
        self.ordinal += 1
        return {
            "drupal_id": measure["const"],
            "ordinal": self.ordinal,
            "name": measure["title"],
        }


class UnitImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Units
    """

    model_class = model.Unit

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
    ]

    def get_source_objects(self):
        """ """
        return list(self.farmos_client.resource.iterate("taxonomy_term", "unit"))

    def normalize_source_object(self, unit):
        """ """
        return {
            "farmos_uuid": UUID(unit["id"]),
            "drupal_id": unit["attributes"]["drupal_internal__tid"],
            "name": unit["attributes"]["name"],
            "description": unit["attributes"]["description"],
        }


class QuantityTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Quantity Types
    """

    model_class = model.QuantityType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
    ]

    def get_source_objects(self):
        """ """
        return list(self.farmos_client.resource.iterate("quantity_type"))

    def normalize_source_object(self, quantity_type):
        """ """
        return {
            "farmos_uuid": UUID(quantity_type["id"]),
            "drupal_id": quantity_type["attributes"]["drupal_internal__id"],
            "name": quantity_type["attributes"]["label"],
            "description": quantity_type["attributes"]["description"],
        }


class LogTypeImporter(FromFarmOS, ToWutta):
    """
    farmOS API → WuttaFarm importer for Log Types
    """

    model_class = model.LogType

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "name",
        "description",
    ]

    def get_source_objects(self):
        """ """
        return list(self.farmos_client.resource.iterate("log_type"))

    def normalize_source_object(self, log_type):
        """ """
        return {
            "farmos_uuid": UUID(log_type["id"]),
            "drupal_id": log_type["attributes"]["drupal_internal__id"],
            "name": log_type["attributes"]["label"],
            "description": log_type["attributes"]["description"],
        }


class LogImporterBase(FromFarmOS, ToWutta):
    """
    Base class for farmOS API → WuttaFarm log importers
    """

    def get_farmos_log_type(self):
        return self.model_class.__wutta_hint__["farmos_log_type"]

    def get_simple_fields(self):
        """ """
        fields = list(super().get_simple_fields())
        # nb. must explicitly declare proxy fields
        fields.extend(
            [
                "farmos_uuid",
                "drupal_id",
                "log_type",
                "message",
                "timestamp",
                "is_movement",
                "is_group_assignment",
                "notes",
                "status",
                "quick",
            ]
        )
        return fields

    def get_supported_fields(self):
        """ """
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "assets",
                "groups",
                "locations",
                "quantities",
                "owners",
            ]
        )
        return fields

    def get_source_objects(self):
        """ """
        log_type = self.get_farmos_log_type()
        return list(self.farmos_client.log.iterate(log_type))

    def normalize_source_object(self, log):
        """ """
        data = self.normal.normalize_farmos_log(log)

        data["farmos_uuid"] = UUID(data.pop("uuid"))
        data["message"] = data.pop("name")
        data["timestamp"] = self.app.make_utc(data["timestamp"])
        data["quick"] = ", ".join(data["quick"]) if data["quick"] else None

        # TODO
        data["log_type"] = self.get_farmos_log_type()

        if "assets" in self.fields:
            data["assets"] = [
                (a["asset_type"], UUID(a["uuid"])) for a in data["assets"]
            ]

        if "groups" in self.fields:
            data["groups"] = [
                (asset["asset_type"], UUID(asset["uuid"])) for asset in data["groups"]
            ]

        if "locations" in self.fields:
            data["locations"] = [
                (asset["asset_type"], UUID(asset["uuid"]))
                for asset in data["locations"]
            ]

        if "quantities" in self.fields:
            data["quantities"] = [UUID(uuid) for uuid in data["quantity_uuids"]]

        if "owners" in self.fields:
            data["owners"] = [UUID(uuid) for uuid in data["owner_uuids"]]

        return data

    def normalize_target_object(self, log):
        data = super().normalize_target_object(log)

        if "assets" in self.fields:
            data["assets"] = [
                (asset.asset_type, asset.farmos_uuid) for asset in log.assets
            ]

        if "groups" in self.fields:
            data["groups"] = [
                (asset.asset_type, asset.farmos_uuid) for asset in log.groups
            ]

        if "locations" in self.fields:
            data["locations"] = [
                (asset.asset_type, asset.farmos_uuid) for asset in log.locations
            ]

        if "quantities" in self.fields:
            data["quantities"] = [qty.farmos_uuid for qty in log.quantities]

        if "owners" in self.fields:
            data["owners"] = [user.farmos_uuid for user in log.owners]

        return data

    def update_target_object(self, log, source_data, target_data=None):
        model = self.app.model
        log = super().update_target_object(log, source_data, target_data)

        if "assets" in self.fields:
            if not target_data or target_data["assets"] != source_data["assets"]:

                for key in source_data["assets"]:
                    asset_type, farmos_uuid = key
                    if not target_data or key not in target_data["assets"]:
                        asset = (
                            self.target_session.query(model.Asset)
                            .filter(model.Asset.asset_type == asset_type)
                            .filter(model.Asset.farmos_uuid == farmos_uuid)
                            .one()
                        )
                        log.assets.append(asset)

                if target_data:
                    for key in target_data["assets"]:
                        asset_type, farmos_uuid = key
                        if key not in source_data["assets"]:
                            asset = (
                                self.target_session.query(model.Asset)
                                .filter(model.Asset.asset_type == asset_type)
                                .filter(model.Asset.farmos_uuid == farmos_uuid)
                                .one()
                            )
                            log.assets.remove(asset)

        if "groups" in self.fields:
            if not target_data or target_data["groups"] != source_data["groups"]:

                for key in source_data["groups"]:
                    asset_type, farmos_uuid = key
                    if not target_data or key not in target_data["groups"]:
                        asset = (
                            self.target_session.query(model.Asset)
                            .filter(model.Asset.asset_type == asset_type)
                            .filter(model.Asset.farmos_uuid == farmos_uuid)
                            .one()
                        )
                        log.groups.append(asset)

                if target_data:
                    for key in target_data["groups"]:
                        asset_type, farmos_uuid = key
                        if key not in source_data["groups"]:
                            asset = (
                                self.target_session.query(model.Asset)
                                .filter(model.Asset.asset_type == asset_type)
                                .filter(model.Asset.farmos_uuid == farmos_uuid)
                                .one()
                            )
                            log.groups.remove(asset)

        if "locations" in self.fields:
            if not target_data or target_data["locations"] != source_data["locations"]:

                for key in source_data["locations"]:
                    asset_type, farmos_uuid = key
                    if not target_data or key not in target_data["locations"]:
                        asset = (
                            self.target_session.query(model.Asset)
                            .filter(model.Asset.asset_type == asset_type)
                            .filter(model.Asset.farmos_uuid == farmos_uuid)
                            .one()
                        )
                        log.locations.append(asset)

                if target_data:
                    for key in target_data["locations"]:
                        asset_type, farmos_uuid = key
                        if key not in source_data["locations"]:
                            asset = (
                                self.target_session.query(model.Asset)
                                .filter(model.Asset.asset_type == asset_type)
                                .filter(model.Asset.farmos_uuid == farmos_uuid)
                                .one()
                            )
                            log.locations.remove(asset)

        if "quantities" in self.fields:
            if (
                not target_data
                or target_data["quantities"] != source_data["quantities"]
            ):

                for farmos_uuid in source_data["quantities"]:
                    if not target_data or farmos_uuid not in target_data["quantities"]:
                        qty = (
                            self.target_session.query(model.Quantity)
                            .filter(model.Quantity.farmos_uuid == farmos_uuid)
                            .one()
                        )
                        log.quantities.append(qty)

                if target_data:
                    for farmos_uuid in target_data["quantities"]:
                        if farmos_uuid not in source_data["quantities"]:
                            qty = (
                                self.target_session.query(model.Quantity)
                                .filter(model.Quantity.farmos_uuid == farmos_uuid)
                                .one()
                            )
                            log.quantities.remove(qty)

        if "owners" in self.fields:
            if not target_data or target_data["owners"] != source_data["owners"]:

                for farmos_uuid in source_data["owners"]:
                    if not target_data or farmos_uuid not in target_data["owners"]:
                        user = (
                            self.target_session.query(model.User)
                            .join(model.WuttaFarmUser)
                            .filter(model.WuttaFarmUser.farmos_uuid == farmos_uuid)
                            .one()
                        )
                        log.owners.append(user)

                if target_data:
                    for farmos_uuid in target_data["owners"]:
                        if farmos_uuid not in source_data["owners"]:
                            user = (
                                self.target_session.query(model.User)
                                .join(model.WuttaFarmUser)
                                .filter(model.WuttaFarmUser.farmos_uuid == farmos_uuid)
                                .one()
                            )
                            log.owners.remove(user)

        return log


class ActivityLogImporter(LogImporterBase):
    """
    farmOS API → WuttaFarm importer for Activity Logs
    """

    model_class = model.ActivityLog


class HarvestLogImporter(LogImporterBase):
    """
    farmOS API → WuttaFarm importer for Harvest Logs
    """

    model_class = model.HarvestLog


class MedicalLogImporter(LogImporterBase):
    """
    farmOS API → WuttaFarm importer for Medical Logs
    """

    model_class = model.MedicalLog

    def get_simple_fields(self):
        """ """
        fields = list(super().get_simple_fields())
        # nb. must explicitly declare proxy fields
        fields.extend(
            [
                "vet",
            ]
        )
        return fields


class ObservationLogImporter(LogImporterBase):
    """
    farmOS API → WuttaFarm importer for Observation Logs
    """

    model_class = model.ObservationLog


class SeedingLogImporter(LogImporterBase):
    """
    farmOS API → WuttaFarm importer for Seeding Logs
    """

    model_class = model.SeedingLog

    def get_simple_fields(self):
        """ """
        fields = list(super().get_simple_fields())
        # nb. must explicitly declare proxy fields
        fields.extend(
            [
                "source",
                "purchase_date",
                "lot_number",
            ]
        )
        return fields

    def normalize_source_object(self, log):
        """ """
        data = super().normalize_source_object(log)
        data.update(
            {
                "source": log["attributes"]["source"],
                "purchase_date": self.normalize_datetime(
                    log["attributes"]["purchase_date"]
                ),
                "lot_number": log["attributes"]["lot_number"],
            }
        )
        return data


class QuantityImporterBase(FromFarmOS, ToWutta):
    """
    Base class for farmOS API → WuttaFarm quantity importers
    """

    def get_farmos_quantity_type(self):
        return self.model_class.__wutta_hint__["farmos_quantity_type"]

    def get_simple_fields(self):
        """ """
        fields = list(super().get_simple_fields())
        # nb. must explicitly declare proxy fields
        fields.extend(
            [
                "farmos_uuid",
                "drupal_id",
                "quantity_type_id",
                "measure_id",
                "value_numerator",
                "value_denominator",
                "units_uuid",
                "label",
            ]
        )
        return fields

    def setup(self):
        super().setup()
        model = self.app.model

        self.quantity_types_by_farmos_uuid = {}
        for quantity_type in self.target_session.query(model.QuantityType):
            if quantity_type.farmos_uuid:
                self.quantity_types_by_farmos_uuid[quantity_type.farmos_uuid] = (
                    quantity_type
                )

        self.units_by_farmos_uuid = {}
        for unit in self.target_session.query(model.Unit):
            if unit.farmos_uuid:
                self.units_by_farmos_uuid[unit.farmos_uuid] = unit

    def get_source_objects(self):
        """ """
        quantity_type = self.get_farmos_quantity_type()
        return list(self.farmos_client.resource.iterate("quantity", quantity_type))

    def get_quantity_type_by_farmos_uuid(self, uuid):
        if hasattr(self, "quantity_types_by_farmos_uuid"):
            return self.quantity_types_by_farmos_uuid.get(UUID(uuid))

        model = self.app.model
        return (
            self.target_session.query(model.QuantityType)
            .filter(model.QuantityType.farmos_uuid == uuid)
            .one()
        )

    def get_unit_by_farmos_uuid(self, uuid):
        if hasattr(self, "units_by_farmos_uuid"):
            return self.units_by_farmos_uuid.get(UUID(uuid))

        model = self.app.model
        return (
            self.target_session.query(model.Unit)
            .filter(model.Unit.farmos_uuid == uuid)
            .one()
        )

    def normalize_source_object(self, quantity):
        """ """
        quantity_type_id = None
        units_uuid = None
        if relationships := quantity.get("relationships"):

            if quantity_type := relationships.get("quantity_type"):
                if quantity_type["data"]:
                    if wf_quantity_type := self.get_quantity_type_by_farmos_uuid(
                        quantity_type["data"]["id"]
                    ):
                        quantity_type_id = wf_quantity_type.drupal_id

            if units := relationships.get("units"):
                if units["data"]:
                    if wf_unit := self.get_unit_by_farmos_uuid(units["data"]["id"]):
                        units_uuid = wf_unit.uuid

        if not quantity_type_id:
            log.warning(
                "missing/invalid quantity_type for farmOS Quantity: %s", quantity
            )
            return None

        if not units_uuid:
            log.warning("missing/invalid units for farmOS Quantity: %s", quantity)
            return None

        value = quantity["attributes"]["value"]

        return {
            "farmos_uuid": UUID(quantity["id"]),
            "drupal_id": quantity["attributes"]["drupal_internal__id"],
            "quantity_type_id": quantity_type_id,
            "measure_id": quantity["attributes"]["measure"],
            "value_numerator": value["numerator"],
            "value_denominator": value["denominator"],
            "units_uuid": units_uuid,
            "label": quantity["attributes"]["label"],
        }


class StandardQuantityImporter(QuantityImporterBase):
    """
    farmOS API → WuttaFarm importer for Standard Quantities
    """

    model_class = model.StandardQuantity

    supported_fields = [
        "farmos_uuid",
        "drupal_id",
        "quantity_type_id",
        "measure_id",
        "value_numerator",
        "value_denominator",
        "units_uuid",
        "label",
    ]


class MaterialQuantityImporter(QuantityImporterBase):
    """
    farmOS API → WuttaFarm importer for Material Quantities
    """

    model_class = model.MaterialQuantity

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "material_types",
            ]
        )
        return fields

    def normalize_source_object(self, quantity):
        """ """
        data = super().normalize_source_object(quantity)

        if "material_types" in self.fields:
            data["material_types"] = [
                UUID(mtype["id"])
                for mtype in quantity["relationships"]["material_type"]["data"]
            ]

        return data

    def normalize_target_object(self, quantity):
        data = super().normalize_target_object(quantity)

        if "material_types" in self.fields:
            data["material_types"] = [
                mtype.farmos_uuid for mtype in quantity.material_types
            ]

        return data

    def update_target_object(self, quantity, source_data, target_data=None):
        model = self.app.model
        quantity = super().update_target_object(quantity, source_data, target_data)

        if "material_types" in self.fields:
            if (
                not target_data
                or target_data["material_types"] != source_data["material_types"]
            ):

                for farmos_uuid in source_data["material_types"]:
                    if (
                        not target_data
                        or farmos_uuid not in target_data["material_types"]
                    ):
                        mtype = (
                            self.target_session.query(model.MaterialType)
                            .filter(model.MaterialType.farmos_uuid == farmos_uuid)
                            .one()
                        )
                        quantity.material_types.append(mtype)

                if target_data:
                    for farmos_uuid in target_data["material_types"]:
                        if farmos_uuid not in source_data["material_types"]:
                            mtype = (
                                self.target_session.query(model.MaterialType)
                                .filter(model.MaterialType.farmos_uuid == farmos_uuid)
                                .one()
                            )
                            quantity.material_types.remove(mtype)

        return quantity



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS Harvest Logs
"""

import colander
from webhelpers2.html import tags

from wuttaweb.forms.schema import WuttaDateTime, WuttaDictEnum
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos import FarmOSMasterView
from wuttafarm.web.grids import (
    ResourceData,
    SimpleSorter,
    StringFilter,
    IntegerFilter,
    DateTimeFilter,
    NullableBooleanFilter,
)
from wuttafarm.web.forms.schema import (
    FarmOSQuantityRefs,
    FarmOSAssetRefs,
    FarmOSRefs,
    LogQuick,
    Notes,
)
from wuttafarm.web.util import render_quantity_objects, render_quantity_object


class LogMasterView(FarmOSMasterView):
    """
    Base class for farmOS Log master views
    """

    farmos_log_type = None
    filterable = True
    sort_on_backend = True

    labels = {
        "name": "Log Name",
        "log_type_name": "Log Type",
        "locations": "Location",
        "quantities": "Quantity",
    }

    grid_columns = [
        "status",
        "drupal_id",
        "timestamp",
        "name",
        "assets",
        "locations",
        "quantities",
        "is_group_assignment",
        "owners",
    ]

    sort_defaults = ("timestamp", "desc")

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
        "status": {"active": True, "verb": "not_equal", "value": "abandoned"},
    }

    form_fields = [
        "name",
        "timestamp",
        "assets",
        "groups",
        "locations",
        "quantities",
        "notes",
        "status",
        "log_type_name",
        "owners",
        "is_movement",
        "is_group_assignment",
        "quick",
        "drupal_id",
    ]

    def get_farmos_api_includes(self):
        return {"log_type", "quantity", "asset", "group", "location", "owner"}

    def get_grid_data(self, **kwargs):
        return ResourceData(
            self.config,
            self.farmos_client,
            f"log--{self.farmos_log_type}",
            include=",".join(self.get_farmos_api_includes()),
            normalizer=self.normalize_log,
        )

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        enum = self.app.enum

        # status
        g.set_enum("status", enum.LOG_STATUS)
        g.set_sorter("status", SimpleSorter("status"))
        g.set_filter(
            "status",
            StringFilter,
            choices=enum.LOG_STATUS,
            verbs=["equal", "not_equal"],
        )

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)
        g.set_sorter("drupal_id", SimpleSorter("drupal_internal__id"))
        g.set_filter("drupal_id", IntegerFilter, path="drupal_internal__id")

        # timestamp
        g.set_renderer("timestamp", "date")
        g.set_link("timestamp")
        g.set_sorter("timestamp", SimpleSorter("timestamp"))
        g.set_filter("timestamp", DateTimeFilter)

        # name
        g.set_link("name")
        g.set_sorter("name", SimpleSorter("name"))
        g.set_filter("name", StringFilter)

        # assets
        g.set_renderer("assets", self.render_assets_for_grid)

        # groups
        g.set_renderer("groups", self.render_assets_for_grid)

        # locations
        g.set_renderer("locations", self.render_assets_for_grid)

        # quantities
        g.set_renderer("quantities", self.render_quantities_for_grid)

        # is_group_assignment
        g.set_renderer("is_group_assignment", "boolean")
        g.set_sorter("is_group_assignment", SimpleSorter("is_group_assignment"))
        g.set_filter("is_group_assignment", NullableBooleanFilter)

        # owners
        g.set_label("owners", "Owner")
        g.set_renderer("owners", self.render_owners_for_grid)

    def render_assets_for_grid(self, log, field, value):
        if not value:
            return ""

        assets = []
        for asset in value:
            if self.farmos_style_grid_links:
                url = self.request.route_url(
                    f"farmos_{asset['asset_type']}_assets.view", uuid=asset["uuid"]
                )
                assets.append(tags.link_to(asset["name"], url))
            else:
                assets.append(asset["name"])
        return ", ".join(assets)

    def render_quantities_for_grid(self, log, field, value):
        if not value:
            return None
        return render_quantity_objects(value)

    def grid_row_class(self, log, data, i):
        if log["status"] == "pending":
            return "has-background-warning"
        if log["status"] == "abandoned":
            return "has-background-danger"
        return None

    def get_instance(self):
        result = self.farmos_client.log.get_id(
            self.farmos_log_type,
            self.request.matchdict["uuid"],
            params={"include": ",".join(self.get_farmos_api_includes())},
        )
        self.raw_json = result
        included = {obj["id"]: obj for obj in result.get("included", [])}
        instance = self.normalize_log(result["data"], included)

        for qty in instance["quantities"]:

            if qty["quantity_type_id"] == "material":
                for mtype in qty["material_types"]:
                    result = self.farmos_client.resource.get_id(
                        "taxonomy_term", "material_type", mtype["uuid"]
                    )
                    mtype["name"] = result["data"]["attributes"]["name"]

            qty["as_text"] = render_quantity_object(qty)

        return instance

    def get_instance_title(self, log):
        return log["name"]

    def normalize_log(self, log, included):
        data = self.normal.normalize_farmos_log(log, included)
        data.update(
            {
                "log_type_name": data["log_type"].get("name"),
            }
        )
        return data

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        enum = self.app.enum
        log = f.model_instance

        # timestamp
        f.set_node("timestamp", WuttaDateTime())
        f.set_widget("timestamp", WuttaDateTimeWidget(self.request))

        # assets
        f.set_node("assets", FarmOSAssetRefs(self.request))

        # groups
        f.set_node("groups", FarmOSAssetRefs(self.request))

        # locations
        f.set_node("locations", FarmOSAssetRefs(self.request))

        # quantities
        f.set_node("quantities", FarmOSQuantityRefs(self.request))

        # is_movement
        f.set_node("is_movement", colander.Boolean())

        # is_group_assignment
        f.set_node("is_group_assignment", colander.Boolean())

        # notes
        f.set_node("notes", Notes())

        # status
        f.set_node("status", WuttaDictEnum(self.request, enum.LOG_STATUS))

        # owners
        if self.creating or self.editing:
            f.remove("owners")  # TODO
        else:
            f.set_node("owners", FarmOSRefs(self.request, "farmos_users"))

        # quick
        f.set_node("quick", LogQuick(self.request))

    def get_xref_buttons(self, log):
        model = self.app.model
        session = self.Session()

        buttons = [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(f"/log/{log['drupal_id']}"),
                target="_blank",
                icon_left="external-link-alt",
            ),
        ]

        if wf_log := (
            session.query(model.Log)
            .filter(model.Log.farmos_uuid == log["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        f"logs_{self.farmos_log_type}.view", uuid=wf_log.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Land Types
"""

from webhelpers2.html import HTML, tags

from wuttafarm.db.model import LandType, LandAsset
from wuttafarm.web.views.assets import AssetTypeMasterView, AssetMasterView
from wuttafarm.web.forms.schema import LandTypeRef


class LandTypeView(AssetTypeMasterView):
    """
    Master view for Land Types
    """

    model_class = LandType
    route_prefix = "land_types"
    url_prefix = "/land-types"

    grid_columns = [
        "name",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "drupal_id",
        "farmos_uuid",
    ]

    has_rows = True
    row_model_class = LandAsset
    rows_viewable = True

    row_grid_columns = [
        "asset_name",
        "is_location",
        "is_fixed",
        "archived",
    ]

    rows_sort_defaults = "asset_name"

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_xref_buttons(self, land_type):
        buttons = super().get_xref_buttons(land_type)

        if land_type.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_land_types.view", uuid=land_type.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons

    def get_row_grid_data(self, land_type):
        model = self.app.model
        session = self.Session()
        return (
            session.query(model.LandAsset)
            .join(model.Asset)
            .filter(model.LandAsset.land_type == land_type)
        )

    def configure_row_grid(self, grid):
        g = grid
        super().configure_row_grid(g)
        model = self.app.model

        # asset_name
        g.set_link("asset_name")
        g.set_sorter("asset_name", model.Asset.asset_name)
        g.set_filter("asset_name", model.Asset.asset_name)

        # is_location
        g.set_renderer("is_location", "boolean")
        g.set_sorter("is_location", model.Asset.is_location)
        g.set_filter("is_location", model.Asset.is_location)

        # is_fixed
        g.set_renderer("is_fixed", "boolean")
        g.set_sorter("is_fixed", model.Asset.is_fixed)
        g.set_filter("is_fixed", model.Asset.is_fixed)

        # archived
        g.set_renderer("archived", "boolean")
        g.set_sorter("archived", model.Asset.archived)
        g.set_filter("archived", model.Asset.archived)

    def get_row_action_url_view(self, land_asset, i):
        return self.request.route_url("land_assets.view", uuid=land_asset.uuid)

    @classmethod
    def defaults(cls, config):
        """ """
        wutta_config = config.registry.settings.get("wutta_config")
        app = wutta_config.get_app()

        if app.is_farmos_mirror():
            cls.creatable = False
            cls.editable = False
            cls.deletable = False

        cls._defaults(config)


class LandAssetView(AssetMasterView):
    """
    Master view for Land Assets
    """

    model_class = LandAsset
    route_prefix = "land_assets"
    url_prefix = "/assets/land"

    farmos_bundle = "land"
    farmos_refurl_path = "/assets/land"

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "asset_name",
        "land_type",
        "parents",
        "archived",
    ]

    form_fields = [
        "asset_name",
        "parents",
        "notes",
        "asset_type",
        "land_type",
        "is_location",
        "is_fixed",
        "archived",
        "drupal_id",
        "farmos_uuid",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model

        # land_type
        g.set_joiner("land_type", lambda q: q.join(model.LandType))
        g.set_sorter("land_type", model.LandType.name)
        g.set_filter("land_type", model.LandType.name, label="Land Type Name")

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        land = f.model_instance

        # land_type
        f.set_node("land_type", LandTypeRef(self.request))


def defaults(config, **kwargs):
    base = globals()

    LandTypeView = kwargs.get("LandTypeView", base["LandTypeView"])
    LandTypeView.defaults(config)

    LandAssetView = kwargs.get("LandAssetView", base["LandAssetView"])
    LandAssetView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Farm Animals
"""

import datetime

import colander
from webhelpers2.html import tags

from wuttaweb.forms.schema import WuttaDateTime, WuttaDictEnum
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos.assets import AssetMasterView
from wuttafarm.web.grids import (
    SimpleSorter,
    StringFilter,
    BooleanFilter,
    NullableBooleanFilter,
    DateTimeFilter,
)
from wuttafarm.web.forms.schema import FarmOSRef


class AnimalView(AssetMasterView):
    """
    Master view for Farm Animals
    """

    model_name = "farmos_animal_assets"
    model_title = "farmOS Animal Asset"
    model_title_plural = "farmOS Animal Assets"

    route_prefix = "farmos_animal_assets"
    url_prefix = "/farmOS/assets/animal"

    farmos_asset_type = "animal"
    farmos_refurl_path = "/assets/animal"

    labels = {
        "animal_type": "Species / Breed",
        "animal_type_name": "Species / Breed",
        "is_sterile": "Sterile",
    }

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "name",
        "produces_eggs",
        "animal_type_name",
        "birthdate",
        "is_sterile",
        "sex",
        "groups",
        "owners",
        "locations",
        "archived",
    ]

    form_fields = [
        "name",
        "animal_type",
        "birthdate",
        "produces_eggs",
        "sex",
        "is_sterile",
        "notes",
        "asset_type_name",
        "owners",
        "locations",
        "groups",
        "archived",
        "thumbnail_url",
        "image_url",
        "thumbnail",
        "image",
    ]

    def get_farmos_api_includes(self):
        includes = super().get_farmos_api_includes()
        includes.update(["animal_type"])
        return includes

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        enum = self.app.enum

        # produces_eggs
        g.set_renderer("produces_eggs", "boolean")
        g.set_sorter("produces_eggs", SimpleSorter("produces_eggs"))
        g.set_filter("produces_eggs", NullableBooleanFilter)

        # animal_type_name
        if self.farmos_style_grid_links:
            g.set_renderer("animal_type_name", self.render_animal_type_for_grid)
        else:
            g.set_link("animal_type_name")
        g.set_sorter("animal_type_name", SimpleSorter("animal_type.name"))
        g.set_filter("animal_type_name", StringFilter, path="animal_type.name")

        # birthdate
        g.set_renderer("birthdate", "date")
        g.set_sorter("birthdate", SimpleSorter("birthdate"))
        g.set_filter("birthdate", DateTimeFilter)

        # sex
        g.set_enum("sex", enum.ANIMAL_SEX)
        g.set_sorter("sex", SimpleSorter("sex"))
        g.set_filter("sex", StringFilter)

        # is_sterile
        g.set_renderer("is_sterile", "boolean")
        g.set_sorter("is_sterile", SimpleSorter("is_sterile"))
        g.set_filter("is_sterile", BooleanFilter)

    def render_animal_type_for_grid(self, animal, field, value):
        uuid = animal["animal_type"]["uuid"]
        url = self.request.route_url("farmos_animal_types.view", uuid=uuid)
        return tags.link_to(value, url)

    def get_instance(self):

        data = super().get_instance()

        if relationships := self.raw_json["data"].get("relationships"):

            # add animal type
            if not data.get("animal_type"):
                if animal_type := relationships.get("animal_type"):
                    if animal_type["data"]:
                        animal_type = self.farmos_client.resource.get_id(
                            "taxonomy_term", "animal_type", animal_type["data"]["id"]
                        )
                        data["animal_type"] = {
                            "uuid": animal_type["data"]["id"],
                            "name": animal_type["data"]["attributes"]["name"],
                        }

        return data

    def normalize_asset(self, animal, included):
        normal = super().normalize_asset(animal, included)

        birthdate = animal["attributes"]["birthdate"]
        if birthdate:
            birthdate = datetime.datetime.fromisoformat(birthdate)
            birthdate = self.app.localtime(birthdate)

        sterile = None
        if self.farmos_4x:
            sterile = animal["attributes"]["is_sterile"]
        else:
            sterile = animal["attributes"]["is_castrated"]

        animal_type_object = None
        if relationships := animal.get("relationships"):

            if animal_type := relationships.get("animal_type"):
                if animal_type := included.get(animal_type["data"]["id"]):
                    animal_type_object = {
                        "uuid": animal_type["id"],
                        "name": animal_type["attributes"]["name"],
                    }

        normal.update(
            {
                "animal_type": animal_type_object,
                "animal_type_uuid": animal_type_object["uuid"],
                "animal_type_name": animal_type_object["name"],
                "birthdate": birthdate,
                "sex": animal["attributes"]["sex"] or colander.null,
                "is_sterile": sterile,
                "produces_eggs": animal["attributes"].get("produces_eggs"),
            }
        )

        return normal

    def get_animal_types(self):
        animal_types = []
        result = self.farmos_client.resource.get(
            "taxonomy_term", "animal_type", params={"sort": "name"}
        )
        for animal_type in result["data"]:
            animal_types.append((animal_type["id"], animal_type["attributes"]["name"]))
        return animal_types

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        enum = self.app.enum
        animal = f.model_instance

        # animal_type
        f.set_node(
            "animal_type",
            FarmOSRef(
                self.request, "farmos_animal_types", values=self.get_animal_types
            ),
        )

        # produces_eggs
        f.set_node("produces_eggs", colander.Boolean())

        # birthdate
        f.set_node("birthdate", WuttaDateTime())
        f.set_widget("birthdate", WuttaDateTimeWidget(self.request))
        f.set_required("birthdate", False)

        # sex
        if not (self.creating or self.editing) and not animal["sex"]:
            pass  # TODO: dict enum widget does not handle null values well
        else:
            f.set_node("sex", WuttaDictEnum(self.request, enum.ANIMAL_SEX))
            f.set_required("sex", False)

        # is_sterile
        f.set_node("is_sterile", colander.Boolean())

    def get_api_payload(self, animal):
        payload = super().get_api_payload(animal)

        birthdate = None
        if animal["birthdate"]:
            birthdate = self.app.localtime(animal["birthdate"]).timestamp()

        attrs = {
            "sex": animal["sex"] or None,
            "is_sterile": animal["is_sterile"],
            "produces_eggs": animal["produces_eggs"],
            "birthdate": birthdate,
        }

        rels = {
            "animal_type": {
                "data": {
                    "id": animal["animal_type"],
                    "type": "taxonomy_term--animal_type",
                }
            }
        }

        payload["attributes"].update(attrs)
        payload.setdefault("relationships", {}).update(rels)
        return payload

    def get_xref_buttons(self, animal):
        buttons = super().get_xref_buttons(animal)

        if self.app.is_farmos_mirror():
            model = self.app.model
            session = self.Session()

            if wf_animal := (
                session.query(model.Asset)
                .filter(model.Asset.farmos_uuid == animal["uuid"])
                .first()
            ):
                buttons.append(
                    self.make_button(
                        f"View {self.app.get_title()} record",
                        primary=True,
                        url=self.request.route_url(
                            "animal_assets.view", uuid=wf_animal.uuid
                        ),
                        icon_left="eye",
                    )
                )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    AnimalView = kwargs.get("AnimalView", base["AnimalView"])
    AnimalView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Asset Types
"""

from wuttafarm.db.model import AssetType
from wuttafarm.web.views import WuttaFarmMasterView


class AssetTypeView(WuttaFarmMasterView):
    """
    Master view for Asset Types
    """

    model_class = AssetType
    route_prefix = "asset_types"
    url_prefix = "/asset-types"

    grid_columns = [
        "name",
        "drupal_id",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "drupal_id",
        "farmos_uuid",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_xref_buttons(self, asset_type):
        buttons = super().get_xref_buttons(asset_type)

        if asset_type.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_asset_types.view", uuid=asset_type.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons

    @classmethod
    def defaults(cls, config):
        """ """
        wutta_config = config.registry.settings.get("wutta_config")
        app = wutta_config.get_app()

        if app.is_farmos_mirror():
            cls.creatable = False
            cls.editable = False
            cls.deletable = False

        cls._defaults(config)


def defaults(config, **kwargs):
    base = globals()

    AssetTypeView = kwargs.get("AssetTypeView", base["AssetTypeView"])
    AssetTypeView.defaults(config)


def includeme(config):
    defaults(config)



"""add LogQuantity

Revision ID: 9e875e5cbdc1
Revises: 74d32b4ec210
Create Date: 2026-02-28 21:55:31.876087

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "9e875e5cbdc1"
down_revision: Union[str, None] = "74d32b4ec210"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # log_quantity
    op.create_table(
        "log_quantity",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("log_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("quantity_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["log_uuid"], ["log.uuid"], name=op.f("fk_log_quantity_log_uuid_log")
        ),
        sa.ForeignKeyConstraint(
            ["quantity_uuid"],
            ["quantity.uuid"],
            name=op.f("fk_log_quantity_quantity_uuid_quantity"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_log_quantity")),
    )
    op.create_table(
        "log_quantity_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "log_uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "quantity_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_log_quantity_version")
        ),
    )
    op.create_index(
        op.f("ix_log_quantity_version_end_transaction_id"),
        "log_quantity_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_quantity_version_operation_type"),
        "log_quantity_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_log_quantity_version_pk_transaction_id",
        "log_quantity_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_log_quantity_version_pk_validity",
        "log_quantity_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_quantity_version_transaction_id"),
        "log_quantity_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # log_quantity
    op.drop_index(
        op.f("ix_log_quantity_version_transaction_id"),
        table_name="log_quantity_version",
    )
    op.drop_index(
        "ix_log_quantity_version_pk_validity", table_name="log_quantity_version"
    )
    op.drop_index(
        "ix_log_quantity_version_pk_transaction_id", table_name="log_quantity_version"
    )
    op.drop_index(
        op.f("ix_log_quantity_version_operation_type"),
        table_name="log_quantity_version",
    )
    op.drop_index(
        op.f("ix_log_quantity_version_end_transaction_id"),
        table_name="log_quantity_version",
    )
    op.drop_table("log_quantity_version")
    op.drop_table("log_quantity")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS Quantity Types
"""

import datetime

import colander
import requests

from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos import FarmOSMasterView
from wuttafarm.web.forms.schema import FarmOSUnitRef, FarmOSRefs
from wuttafarm.web.grids import ResourceData


class QuantityTypeView(FarmOSMasterView):
    """
    View for farmOS Quantity Types
    """

    model_name = "farmos_quantity_type"
    model_title = "farmOS Quantity Type"
    model_title_plural = "farmOS Quantity Types"

    route_prefix = "farmos_quantity_types"
    url_prefix = "/farmOS/quantity-types"

    grid_columns = [
        "label",
        "description",
    ]

    sort_defaults = "label"

    form_fields = [
        "label",
        "description",
    ]

    def get_grid_data(self, columns=None, session=None):
        result = self.farmos_client.resource.get("quantity_type")
        return [self.normalize_quantity_type(t) for t in result["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # label
        g.set_link("label")
        g.set_searchable("label")

        # description
        g.set_searchable("description")

    def get_instance(self):
        result = self.farmos_client.resource.get_id(
            "quantity_type", "quantity_type", self.request.matchdict["uuid"]
        )
        self.raw_json = result
        return self.normalize_quantity_type(result["data"])

    def get_instance_title(self, quantity_type):
        return quantity_type["label"]

    def normalize_quantity_type(self, quantity_type):
        return {
            "uuid": quantity_type["id"],
            "drupal_id": quantity_type["attributes"]["drupal_internal__id"],
            "label": quantity_type["attributes"]["label"],
            "description": quantity_type["attributes"]["description"],
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # description
        f.set_widget("description", "notes")

    def get_xref_buttons(self, quantity_type):
        model = self.app.model
        session = self.Session()
        buttons = []

        if wf_quantity_type := (
            session.query(model.QuantityType)
            .filter(model.QuantityType.farmos_uuid == quantity_type["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "quantity_types.view", uuid=wf_quantity_type.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class QuantityMasterView(FarmOSMasterView):
    """
    Base class for Quantity views
    """

    farmos_quantity_type = None

    grid_columns = [
        "drupal_id",
        "as_text",
        "measure",
        "value",
        "unit",
        "label",
    ]

    sort_defaults = ("drupal_id", "desc")

    form_fields = [
        "quantity_type_name",
        "measure",
        "value",
        "units",
        "label",
        "created",
        "changed",
    ]

    def get_farmos_api_includes(self):
        return {"units"}

    def get_grid_data(self, **kwargs):
        return ResourceData(
            self.config,
            self.farmos_client,
            f"quantity--{self.farmos_quantity_type}",
            include=",".join(self.get_farmos_api_includes()),
            normalizer=self.normalize_quantity,
        )

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)

        # as_text
        g.set_renderer("as_text", self.render_as_text_for_grid)
        g.set_link("as_text")

        # measure
        g.set_renderer("measure", self.render_measure_for_grid)

        # value
        g.set_renderer("value", self.render_value_for_grid)

        # unit
        g.set_renderer("unit", self.render_unit_for_grid)

        # changed
        g.set_renderer("changed", "datetime")

    def render_as_text_for_grid(self, qty, field, value):
        measure = qty["measure"].capitalize()
        value = qty["value"]["decimal"]
        units = qty["unit"]["name"] if qty["unit"] else "??"
        return f"( {measure} ) {value} {units}"

    def render_measure_for_grid(self, qty, field, value):
        return qty["measure"].capitalize()

    def render_unit_for_grid(self, qty, field, value):
        unit = qty[field]
        if not unit:
            return ""
        return unit["name"]

    def render_value_for_grid(self, qty, field, value):
        return qty["value"]["decimal"]

    def get_instance(self):
        # TODO: this pattern should be repeated for other views
        try:
            result = self.farmos_client.resource.get_id(
                "quantity",
                self.farmos_quantity_type,
                self.request.matchdict["uuid"],
                params={"include": self.get_farmos_api_includes()},
            )
        except requests.HTTPError as exc:
            if exc.response.status_code == 404:
                raise self.notfound()
            raise

        self.raw_json = result

        included = {obj["id"]: obj for obj in result.get("included", [])}
        assert included
        data = self.normalize_quantity(result["data"], included)

        if relationships := result["data"].get("relationships"):

            # add units
            if units := relationships.get("units"):
                if units["data"]:
                    unit = self.farmos_client.resource.get_id(
                        "taxonomy_term", "unit", units["data"]["id"]
                    )
                    data["units"] = {
                        "uuid": unit["data"]["id"],
                        "name": unit["data"]["attributes"]["name"],
                    }

        return data

    def get_instance_title(self, quantity):
        return quantity["value"]

    def normalize_quantity(self, quantity, included={}):

        if created := quantity["attributes"]["created"]:
            created = datetime.datetime.fromisoformat(created)
            created = self.app.localtime(created)

        if changed := quantity["attributes"]["changed"]:
            changed = datetime.datetime.fromisoformat(changed)
            changed = self.app.localtime(changed)

        quantity_type_object = None
        quantity_type_uuid = None
        unit_object = None
        unit_uuid = None
        if relationships := quantity["relationships"]:

            if quantity_type := relationships["quantity_type"]["data"]:
                quantity_type_uuid = quantity_type["id"]
                quantity_type_object = {
                    "uuid": quantity_type_uuid,
                    "type": "quantity_type--quantity_type",
                }

            if unit := relationships["units"]["data"]:
                unit_uuid = unit["id"]
                if unit := included.get(unit_uuid):
                    unit_object = {
                        "uuid": unit_uuid,
                        "type": "taxonomy_term--unit",
                        "name": unit["attributes"]["name"],
                    }

        return {
            "uuid": quantity["id"],
            "drupal_id": quantity["attributes"]["drupal_internal__id"],
            "quantity_type": quantity_type_object,
            "quantity_type_uuid": quantity_type_uuid,
            "measure": quantity["attributes"]["measure"],
            "value": quantity["attributes"]["value"],
            "unit": unit_object,
            "unit_uuid": unit_uuid,
            "label": quantity["attributes"]["label"] or colander.null,
            "created": created,
            "changed": changed,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # quantity_type_name
        f.set_label("quantity_type_name", "Quantity Type")
        f.set_readonly("quantity_type_name")
        f.set_default("quantity_type_name", self.farmos_quantity_type.capitalize())

        # created
        f.set_node("created", WuttaDateTime(self.request))
        f.set_widget("created", WuttaDateTimeWidget(self.request))

        # changed
        f.set_node("changed", WuttaDateTime(self.request))
        f.set_widget("changed", WuttaDateTimeWidget(self.request))

        # units
        f.set_node("units", FarmOSUnitRef())


class StandardQuantityView(QuantityMasterView):
    """
    View for farmOS Standard Quantities
    """

    model_name = "farmos_standard_quantity"
    model_title = "farmOS Standard Quantity"
    model_title_plural = "farmOS Standard Quantities"

    route_prefix = "farmos_quantities_standard"
    url_prefix = "/farmOS/quantities/standard"

    farmos_quantity_type = "standard"
    farmos_refurl_path = "/log-quantities/standard"

    def get_xref_buttons(self, standard_quantity):
        model = self.app.model
        session = self.Session()
        buttons = []

        if wf_standard_quantity := (
            session.query(model.StandardQuantity)
            .join(model.Quantity)
            .filter(model.Quantity.farmos_uuid == standard_quantity["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "quantities_standard.view", uuid=wf_standard_quantity.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class MaterialQuantityView(QuantityMasterView):
    """
    View for farmOS Material Quantities
    """

    model_name = "farmos_material_quantity"
    model_title = "farmOS Material Quantity"
    model_title_plural = "farmOS Material Quantities"

    route_prefix = "farmos_quantities_material"
    url_prefix = "/farmOS/quantities/material"

    farmos_quantity_type = "material"
    farmos_refurl_path = "/log-quantities/material"

    def get_farmos_api_includes(self):
        includes = super().get_farmos_api_includes()
        includes.update({"material_type"})
        return includes

    def normalize_quantity(self, quantity, included={}):
        normal = super().normalize_quantity(quantity, included)

        material_type_objects = []
        material_type_uuids = []
        if relationships := quantity["relationships"]:

            if material_types := relationships["material_type"]["data"]:
                for mtype in material_types:
                    uuid = mtype["id"]
                    material_type_uuids.append(uuid)
                    material_type = {
                        "uuid": uuid,
                        "type": mtype["type"],
                    }
                    if mtype := included.get(uuid):
                        material_type.update(
                            {
                                "name": mtype["attributes"]["name"],
                            }
                        )
                    material_type_objects.append(material_type)

        normal.update(
            {
                "material_types": material_type_objects,
                "material_type_uuids": material_type_uuids,
            }
        )
        return normal

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # material_types
        f.fields.insert_before("measure", "material_types")
        f.set_node("material_types", FarmOSRefs(self.request, "farmos_material_types"))

    def get_xref_buttons(self, material_quantity):
        model = self.app.model
        session = self.Session()
        buttons = []

        if wf_material_quantity := (
            session.query(model.MaterialQuantity)
            .join(model.Quantity)
            .filter(model.Quantity.farmos_uuid == material_quantity["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "quantities_material.view", uuid=wf_material_quantity.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    QuantityTypeView = kwargs.get("QuantityTypeView", base["QuantityTypeView"])
    QuantityTypeView.defaults(config)

    StandardQuantityView = kwargs.get(
        "StandardQuantityView", base["StandardQuantityView"]
    )
    StandardQuantityView.defaults(config)

    MaterialQuantityView = kwargs.get(
        "MaterialQuantityView", base["MaterialQuantityView"]
    )
    MaterialQuantityView.defaults(config)


def includeme(config):
    defaults(config)



"""remove AnimalType.changed

Revision ID: b8cd4a8f981f
Revises: aecfd9175624
Create Date: 2026-02-17 18:11:06.110003

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "b8cd4a8f981f"
down_revision: Union[str, None] = "aecfd9175624"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # animal_type
    op.drop_column("animal_type", "changed")


def downgrade() -> None:

    # animal_type
    op.add_column(
        "animal_type",
        sa.Column(
            "changed", postgresql.TIMESTAMP(), autoincrement=False, nullable=True
        ),
    )



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model definition for Asset Types
"""

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.associationproxy import association_proxy

from wuttjamaican.db import model


class AssetType(model.Base):
    """
    Represents an "asset type" from farmOS
    """

    __tablename__ = "asset_type"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Asset Type",
        "model_title_plural": "Asset Types",
    }

    uuid = model.uuid_column()

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        unique=True,
        doc="""
        Name of the asset type.
        """,
    )

    description = sa.Column(
        sa.String(length=255),
        nullable=True,
        doc="""
        Description for the asset type.
        """,
    )

    farmos_uuid = sa.Column(
        model.UUID(),
        nullable=True,
        unique=True,
        doc="""
        UUID for the asset type within farmOS.
        """,
    )

    drupal_id = sa.Column(
        sa.String(length=50),
        nullable=True,
        unique=True,
        doc="""
        Drupal internal ID for the asset type.
        """,
    )

    def __str__(self):
        return self.name or ""


class Asset(model.Base):
    """
    Represents an asset (of any kind) from farmOS.
    """

    __tablename__ = "asset"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Asset",
        "model_title_plural": "All Assets",
    }

    uuid = model.uuid_column()

    farmos_uuid = sa.Column(
        model.UUID(),
        nullable=True,
        unique=True,
        doc="""
        UUID for the asset within farmOS.
        """,
    )

    drupal_id = sa.Column(
        sa.Integer(),
        nullable=True,
        unique=True,
        doc="""
        Drupal internal ID for the asset.
        """,
    )

    asset_type = sa.Column(
        sa.String(length=100), sa.ForeignKey("asset_type.drupal_id"), nullable=False
    )

    asset_name = sa.Column(
        sa.String(length=100),
        nullable=False,
        doc="""
        Name of the asset.
        """,
    )

    is_location = sa.Column(
        sa.Boolean(),
        nullable=False,
        default=False,
        doc="""
        Whether the asset should be considered a location.
        """,
    )

    is_fixed = sa.Column(
        sa.Boolean(),
        nullable=False,
        default=False,
        doc="""
        Whether the asset's location is fixed.
        """,
    )

    notes = sa.Column(
        sa.Text(),
        nullable=True,
        doc="""
        Notes for the asset.
        """,
    )

    thumbnail_url = sa.Column(
        sa.String(length=255),
        nullable=True,
        doc="""
        Optional thumbnail URL for the asset.
        """,
    )

    image_url = sa.Column(
        sa.String(length=255),
        nullable=True,
        doc="""
        Optional image URL for the asset.
        """,
    )

    archived = sa.Column(
        sa.Boolean(),
        nullable=False,
        default=False,
        doc="""
        Whether the asset is archived.
        """,
    )

    _parents = orm.relationship(
        "AssetParent",
        foreign_keys="AssetParent.asset_uuid",
        back_populates="asset",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
    )

    parents = association_proxy(
        "_parents",
        "parent",
        creator=lambda parent: AssetParent(parent=parent),
    )

    _owners = orm.relationship(
        "AssetOwner",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
        back_populates="asset",
    )

    owners = association_proxy(
        "_owners",
        "user",
        creator=lambda user: AssetOwner(user=user),
    )

    def __str__(self):
        return self.asset_name or ""


class AssetMixin:

    uuid = model.uuid_fk_column("asset.uuid", nullable=False, primary_key=True)

    @declared_attr
    def asset(cls):
        return orm.relationship(
            Asset,
            single_parent=True,
            cascade="all, delete-orphan",
            cascade_backrefs=False,
        )

    def __str__(self):
        return self.asset_name or ""


def add_asset_proxies(subclass):
    Asset.make_proxy(subclass, "asset", "farmos_uuid")
    Asset.make_proxy(subclass, "asset", "drupal_id")
    Asset.make_proxy(subclass, "asset", "asset_type")
    Asset.make_proxy(subclass, "asset", "asset_name")
    Asset.make_proxy(subclass, "asset", "is_location")
    Asset.make_proxy(subclass, "asset", "is_fixed")
    Asset.make_proxy(subclass, "asset", "notes")
    Asset.make_proxy(subclass, "asset", "thumbnail_url")
    Asset.make_proxy(subclass, "asset", "image_url")
    Asset.make_proxy(subclass, "asset", "archived")
    Asset.make_proxy(subclass, "asset", "parents")
    Asset.make_proxy(subclass, "asset", "owners")


class EggMixin:

    produces_eggs = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
        Whether the group asset produces eggs (i.e. it should be
        available in the egg harvest form).
        """,
    )


class AssetParent(model.Base):
    """
    Represents an "asset's parent relationship" from farmOS.
    """

    __tablename__ = "asset_parent"
    __versioned__ = {}

    uuid = model.uuid_column()

    asset_uuid = model.uuid_fk_column("asset.uuid", nullable=False)

    asset = orm.relationship(
        Asset,
        foreign_keys=asset_uuid,
    )

    parent_uuid = model.uuid_fk_column("asset.uuid", nullable=False)

    parent = orm.relationship(
        Asset,
        foreign_keys=parent_uuid,
    )


class AssetOwner(model.Base):
    """
    Represents a "asset's owner relationship" from farmOS.
    """

    __tablename__ = "asset_owner"
    __versioned__ = {}

    uuid = model.uuid_column()

    asset_uuid = model.uuid_fk_column("asset.uuid", nullable=False)
    asset = orm.relationship(
        Asset,
        foreign_keys=asset_uuid,
        back_populates="_owners",
    )

    user_uuid = model.uuid_fk_column("user.uuid", nullable=False)
    user = orm.relationship(
        model.User,
        foreign_keys=user_uuid,
    )
